# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contextual component and app test."""

import inspect
from typing import Any
import unittest
import weakref

from langfun.core import component as lf
import pyglove as pg


class ComponentContextTest(unittest.TestCase):
  """Tests for ComponentContext."""

  def test_override(self):
    class A(lf.Component):
      x: int

    a = A(x=1)
    with a.override(x=2, y=1):
      self.assertEqual(a.x, 2)

      # `y`` is not an attribute of `A`.
      with self.assertRaises(AttributeError):
        _ = a.y

  def test_context(self):
    class A(lf.Component):
      x: int
      y: int = lf.contextual()
      z: int = lf.contextual(default=-1)

    with self.assertRaisesRegex(TypeError, '.* missing 1 required argument'):
      _ = A()

    a = A(x=1)
    with self.assertRaisesRegex(
        AttributeError, 'p'
    ):
      _ = a.p

    with self.assertRaisesRegex(
        AttributeError, '.* is not found under its context'
    ):
      _ = a.y

    with lf.context(y=1):
      self.assertEqual(a.y, 1)

    with self.assertRaisesRegex(
        AttributeError, '.* is not found under its context'
    ):
      _ = a.y

    # Use contextual default if it's not provided.
    self.assertEqual(a.z, -1)

    a1 = A(x=1, y=2)
    self.assertEqual(a1.x, 1)
    self.assertEqual(a1.y, 2)
    self.assertEqual(a1.z, -1)

    with lf.context(x=3, y=3, z=3):
      # Member attributes take precedence over `lf.context`.
      self.assertEqual(a1.x, 1)
      self.assertEqual(a1.y, 2)

      # Override attributes take precedence over member attribute.
      with a1.override(y=3):
        self.assertEqual(a1.y, 3)
        with a1.override(y=4):
          self.assertEqual(a1.y, 4)
        self.assertEqual(a1.y, 3)
      self.assertEqual(a1.y, 2)

      # `lf.context` takes precedence over contextual default.
      self.assertEqual(a1.z, 3)

      # Test nested contextual override with override_attrs=True (default).
      with lf.context(y=4, z=4, override_attrs=True):

        # Member attribute is not overriden as current scope does not override
        # `x``.
        self.assertEqual(a1.x, 1)

        # Member attribute is overriden.
        self.assertEqual(a1.y, 4)

        # `lf.Component.override` takes precedence over
        # `lf.context(override_attrs=True)`.
        with a1.override(y=3):
          self.assertEqual(a1.y, 3)
        self.assertEqual(a1.y, 4)

        # Member default is overriden.
        self.assertEqual(a1.z, 4)

      self.assertEqual(a1.y, 2)
      self.assertEqual(a1.z, 3)

    self.assertEqual(a1.y, 2)
    self.assertEqual(a1.z, -1)

  def test_context_cascade(self):
    class A(lf.Component):
      x: int
      y: int = lf.contextual()
      z: int = lf.contextual(default=-1)

    a = A(1, 2)
    self.assertEqual(a.x, 1)
    self.assertEqual(a.y, 2)
    self.assertEqual(a.z, -1)

    with lf.context(x=3, y=3, z=3, cascade=True):
      self.assertEqual(a.x, 1)
      self.assertEqual(a.y, 2)
      self.assertEqual(a.z, 3)

      # Outter `lf.force_context` takes precedence
      # over inner `lf.force_context`.
      with lf.context(y=4, z=4, cascade=True):
        self.assertEqual(a.x, 1)
        self.assertEqual(a.y, 2)
        self.assertEqual(a.z, 3)

      with lf.context(y=4, z=4, override_attrs=True):
        self.assertEqual(a.x, 1)
        self.assertEqual(a.y, 2)
        self.assertEqual(a.z, 3)

    self.assertEqual(a.x, 1)
    self.assertEqual(a.y, 2)
    self.assertEqual(a.z, -1)

    with lf.context(x=3, y=3, z=3, cascade=True, override_attrs=True):
      self.assertEqual(a.x, 3)
      self.assertEqual(a.y, 3)
      self.assertEqual(a.z, 3)

      with lf.context(y=4, z=4, override_attrs=True):
        self.assertEqual(a.x, 3)
        self.assertEqual(a.y, 3)
        self.assertEqual(a.z, 3)

    self.assertEqual(a.x, 1)
    self.assertEqual(a.y, 2)
    self.assertEqual(a.z, -1)

  def test_sym_inferred(self):
    class A(lf.Component):
      x: int = 1
      y: int = lf.contextual()

    a = A()
    with self.assertRaisesRegex(
        AttributeError, '.* is not found under its context'):
      _ = a.sym_inferred('y')
    self.assertIsNone(a.sym_inferred('y', default=None))

    with self.assertRaises(AttributeError):
      _ = a.sym_inferred('z')
    self.assertIsNone(a.sym_inferred('z', default=None))

  def test_symbolic_assignment(self):
    class A(lf.Component):
      x: int = 1

      def _on_bound(self):
        super()._on_bound()
        self.y = self.x * 2

    a = A()
    self.assertEqual(a.y, 2)
    a.x = 2
    self.assertEqual(a.y, 4)

  def test_symbolic_eq_ne_hash(self):
    class A(lf.Component):
      x: int = 1

    a = A()
    self.assertNotEqual(a, A())
    self.assertTrue(pg.eq(a, A()))
    self.assertFalse(pg.eq(a, A(2)))
    self.assertNotEqual(hash(a), hash(A()))
    self.assertEqual(pg.hash(a), pg.hash(A()))

  def test_weak_ref(self):
    class A(lf.Component):
      x: int = 1

    a = A()
    self.assertIsNotNone(weakref.ref(a))


class ContextualAttributeTest(unittest.TestCase):
  """Tests for Component."""

  def test_contextualibute_access(self):

    class A(lf.Component):
      x: int
      y: int = lf.contextual()

    # Not okay: `A.x` is required.
    with self.assertRaisesRegex(TypeError, 'missing 1 required argument'):
      _ = A()

    # Okay: `A.y` is contextual.
    a = A(1)

    # `a.y` is not yet available from the context.
    with self.assertRaises(AttributeError):
      _ = a.y

    class B(lf.Component):
      # Attributes with annotation will be treated as symbolic fields.
      p: int
      q: A = A(2)
      z: int = lf.contextual()

    class C(lf.Component):
      a: int

      # Attributes of lf type without annotation will also be
      # treated as symbolic fields.
      b = B(2)

      y: int = 1

      # Attributes of non-componentl type without annotation will not
      # be treated as symbolic fields.
      z = 2

    c = C(1)
    b = c.b
    a = b.q

    # Test symbolic attributes declared from C.
    self.assertTrue(c.sym_hasattr('a'))
    self.assertTrue(c.sym_hasattr('b'))
    self.assertTrue(c.sym_hasattr('y'))
    self.assertFalse(c.sym_hasattr('z'))

    # Contextual access to c.y from a.
    self.assertEqual(a.y, 1)
    self.assertEqual(b.z, 2)

    # 'y' is not defined as an attribute in 'B'.
    with self.assertRaises(AttributeError):
      _ = b.y

    c.rebind(y=2)
    self.assertEqual(c.y, 2)
    self.assertEqual(a.y, 2)

    c.z = 3
    self.assertEqual(c.z, 3)
    self.assertEqual(b.z, 3)

  def test_to_html(self):
    class A(lf.Component):
      x: int = 1
      y: int = lf.contextual()

    def assert_content(html, expected):
      expected = inspect.cleandoc(expected).strip()
      actual = html.content.strip()
      if actual != expected:
        print(actual)
      self.assertEqual(actual.strip(), expected)

    self.assertIn(
        inspect.cleandoc(
            """
            .contextual-attribute {
              color: purple;
            }
            .unavailable-contextual {
              color: gray;
              font-style: italic;
            }
            """
        ),
        A().to_html().style_section,
    )

    assert_content(
        A().to_html(enable_summary_tooltip=False),
        """
        <details open class="pyglove a"><summary><div class="summary-title">A(...)</div></summary><div class="complex-value a"><details open class="pyglove int"><summary><div class="summary-name">x<span class="tooltip">x</span></div><div class="summary-title">int</div></summary><span class="simple-value int">1</span></details><details open class="pyglove contextual-attribute"><summary><div class="summary-name">y<span class="tooltip">y</span></div><div class="summary-title">ContextualAttribute(...)</div></summary><div class="unavailable-contextual">(not available)</div></details></div></details>
        """
    )

    class B(lf.Component):
      z: Any
      y: int = 2

    b = B(A())
    assert_content(
        b.z.to_html(enable_summary_tooltip=False),
        """
        <details open class="pyglove a"><summary><div class="summary-title">A(...)</div></summary><div class="complex-value a"><details open class="pyglove int"><summary><div class="summary-name">x<span class="tooltip">x</span></div><div class="summary-title">int</div></summary><span class="simple-value int">1</span></details><details open class="pyglove contextual-attribute"><summary><div class="summary-name">y<span class="tooltip">y</span></div><div class="summary-title">ContextualAttribute(...)</div></summary><span class="simple-value int">2</span></details></div></details>
        """
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import unittest
import langfun.core as lf
from langfun.core import modalities as lf_modalities
from langfun.core.data.conversion import openai  # pylint: disable=unused-import


image_content = (
    b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x18\x00\x00\x00\x18\x04'
    b'\x03\x00\x00\x00\x12Y \xcb\x00\x00\x00\x18PLTE\x00\x00'
    b'\x00fff_chaag_cg_ch^ci_ciC\xedb\x94\x00\x00\x00\x08tRNS'
    b'\x00\n\x9f*\xd4\xff_\xf4\xe4\x8b\xf3a\x00\x00\x00>IDATx'
    b'\x01c \x05\x08)"\xd8\xcc\xae!\x06pNz\x88k\x19\\Q\xa8"\x10'
    b'\xc1\x14\x95\x01%\xc1\n\xa143Ta\xa8"D-\x84\x03QM\x98\xc3'
    b'\x1a\x1a\x1a@5\x0e\x04\xa0q\x88\x05\x00\x07\xf8\x18\xf9'
    b'\xdao\xd0|\x00\x00\x00\x00IEND\xaeB`\x82'
)


class OpenAIConversionTest(unittest.TestCase):

  def test_as_format_with_role(self):
    self.assertEqual(
        lf.UserMessage('hi').as_format('openai'),
        {
            'role': 'user',
            'content': [{'type': 'text', 'text': 'hi'}],
        },
    )
    self.assertEqual(
        lf.AIMessage('hi').as_format('openai'),
        {
            'role': 'assistant',
            'content': [{'type': 'text', 'text': 'hi'}],
        },
    )
    self.assertEqual(
        lf.SystemMessage('hi').as_format('openai'),
        {
            'role': 'system',
            'content': [{'type': 'text', 'text': 'hi'}],
        },
    )

  def test_as_format_with_image(self):
    self.assertEqual(
        lf.Template(
            'What is this {{image}}?',
            image=lf_modalities.Image.from_bytes(image_content)
        ).render().as_format('openai'),
        {
            'role': 'user',
            'content': [
                {
                    'type': 'text',
                    'text': 'What is this'
                },
                {
                    'type': 'image_url',
                    'image_url': {
                        'url': (
                            'data:image/png;base64,'
                            + base64.b64encode(image_content).decode('utf-8')
                        )
                    }
                },
                {
                    'type': 'text',
                    'text': '?'
                }
            ],
        },
    )

  def test_as_format_with_chunk_preprocessor(self):
    self.assertEqual(
        lf.Template(
            'What is this {{image}}?',
            image=lf_modalities.Image.from_bytes(image_content)
        ).render().as_openai_format(
            chunk_preprocessor=lambda x: x if isinstance(x, str) else None
        ),
        {
            'role': 'user',
            'content': [
                {
                    'type': 'text',
                    'text': 'What is this'
                },
                {
                    'type': 'text',
                    'text': '?'
                }
            ],
        },
    )

  def test_from_value_with_simple_text(self):
    self.assertEqual(
        lf.Message.from_value(
            {
                'content': 'this is a text',
            },
            format='openai',
        ),
        lf.AIMessage('this is a text'),
    )

  def test_from_value_with_role(self):
    self.assertEqual(
        lf.Message.from_value(
            {
                'role': 'user',
                'content': [{'type': 'text', 'text': 'hi'}],
            },
            format='openai',
        ),
        lf.UserMessage('hi'),
    )
    self.assertEqual(
        lf.Message.from_value(
            {
                'role': 'assistant',
                'content': [{'type': 'text', 'text': 'hi'}],
            },
            format='openai',
        ),
        lf.AIMessage('hi'),
    )
    self.assertEqual(
        lf.Message.from_value(
            {
                'role': 'system',
                'content': [{'type': 'text', 'text': 'hi'}],
            },
            format='openai',
        ),
        lf.SystemMessage('hi'),
    )
    with self.assertRaisesRegex(ValueError, 'Unsupported role: .*'):
      lf.Message.from_value(
          {
              'role': 'function',
              'content': [{'type': 'text', 'text': 'hi'}],
          },
          format='openai',
      )

  def test_from_value_with_image(self):
    m = lf.Message.from_openai_format(
        lf.Template(
            'What is this {{image}}?',
            image=lf_modalities.Image.from_bytes(image_content)
        ).render().as_format('openai'),
    )
    self.assertEqual(m.text, 'What is this <<[[obj0]]>> ?')
    self.assertIsInstance(m.obj0, lf_modalities.Image)
    self.assertEqual(m.obj0.mime_type, 'image/png')
    self.assertEqual(m.obj0.to_bytes(), image_content)


if __name__ == '__main__':
  unittest.main()



# Copyright 2025 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import unittest
import langfun.core as lf
from langfun.core import modalities as lf_modalities
from langfun.core.data.conversion import anthropic  # pylint: disable=unused-import


image_content = (
    b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x18\x00\x00\x00\x18\x04'
    b'\x03\x00\x00\x00\x12Y \xcb\x00\x00\x00\x18PLTE\x00\x00'
    b'\x00fff_chaag_cg_ch^ci_ciC\xedb\x94\x00\x00\x00\x08tRNS'
    b'\x00\n\x9f*\xd4\xff_\xf4\xe4\x8b\xf3a\x00\x00\x00>IDATx'
    b'\x01c \x05\x08)"\xd8\xcc\xae!\x06pNz\x88k\x19\\Q\xa8"\x10'
    b'\xc1\x14\x95\x01%\xc1\n\xa143Ta\xa8"D-\x84\x03QM\x98\xc3'
    b'\x1a\x1a\x1a@5\x0e\x04\xa0q\x88\x05\x00\x07\xf8\x18\xf9'
    b'\xdao\xd0|\x00\x00\x00\x00IEND\xaeB`\x82'
)

pdf_content = (
    b'%PDF-1.1\n%\xc2\xa5\xc2\xb1\xc3\xab\n\n1 0 obj\n'
    b'<< /Type /Catalog\n     /Pages 2 0 R\n  >>\nendobj\n\n2 0 obj\n '
    b'<< /Type /Pages\n     /Kids [3 0 R]\n     '
    b'/Count 1\n     /MediaBox [0 0 300 144]\n  '
    b'>>\nendobj\n\n3 0 obj\n  '
    b'<<  /Type /Page\n      /Parent 2 0 R\n      /Resources\n       '
    b'<< /Font\n'
    b'<< /F1\n'
    b'<< /Type /Font\n'
    b'/Subtype /Type1\n'
    b'/BaseFont /Times-Roman\n'
    b'>>\n>>\n>>\n      '
    b'/Contents 4 0 R\n  >>\nendobj\n\n4 0 obj\n  '
    b'<< /Length 55 >>\nstream\n  BT\n    /F1 18 Tf\n    0 0 Td\n    '
    b'(Hello World) Tj\n  ET\nendstream\nendobj\n\nxref\n0 5\n0000000000 '
    b'65535 f \n0000000018 00000 n \n0000000077 00000 n \n0000000178 00000 n '
    b'\n0000000457 00000 n \ntrailer\n  <<  /Root 1 0 R\n      /Size 5\n  '
    b'>>\nstartxref\n565\n%%EOF\n'
)


class AnthropicConversionTest(unittest.TestCase):

  def test_as_format_with_role(self):
    self.assertEqual(
        lf.UserMessage('hi').as_format('anthropic'),
        {
            'role': 'user',
            'content': [{'type': 'text', 'text': 'hi'}],
        },
    )
    self.assertEqual(
        lf.AIMessage('hi').as_format('anthropic'),
        {
            'role': 'assistant',
            'content': [{'type': 'text', 'text': 'hi'}],
        },
    )
    self.assertEqual(
        lf.SystemMessage('hi').as_format('anthropic'),
        {
            'role': 'system',
            'content': [{'type': 'text', 'text': 'hi'}],
        },
    )

  def test_as_format_with_image(self):
    self.assertEqual(
        lf.Template(
            'What are the common words from {{image}} and {{pdf}}?',
            image=lf_modalities.Image.from_bytes(image_content),
            pdf=lf_modalities.PDF.from_bytes(pdf_content),
        ).render().as_anthropic_format(),
        {
            'role': 'user',
            'content': [
                {
                    'type': 'text',
                    'text': 'What are the common words from'
                },
                {
                    'type': 'image',
                    'source': {
                        'type': 'base64',
                        'media_type': 'image/png',
                        'data': base64.b64encode(
                            image_content
                        ).decode('utf-8'),
                    }
                },
                {
                    'type': 'text',
                    'text': 'and'
                },
                {
                    'type': 'document',
                    'source': {
                        'type': 'base64',
                        'media_type': 'application/pdf',
                        'data': base64.b64encode(
                            pdf_content
                        ).decode('utf-8'),
                    }
                },
                {
                    'type': 'text',
                    'text': '?'
                },
            ],
        },
    )

  def test_as_format_with_chunk_preprocessor(self):
    self.assertEqual(
        lf.Template(
            'What is this {{image}}?',
            image=lf_modalities.Image.from_bytes(image_content)
        ).render().as_format(
            'anthropic',
            chunk_preprocessor=lambda x: x if isinstance(x, str) else None
        ),
        {
            'role': 'user',
            'content': [
                {
                    'type': 'text', 'text': 'What is this'
                },
                {
                    'type': 'text', 'text': '?'
                }
            ],
        },
    )

  def test_from_value_with_simple_text(self):
    self.assertEqual(
        lf.Message.from_value(
            {
                'content': [{'type': 'text', 'text': 'this is a text'}],
            },
            format='anthropic',
        ),
        lf.AIMessage('this is a text'),
    )

  def test_from_value_with_role(self):
    self.assertEqual(
        lf.Message.from_value(
            {
                'role': 'user',
                'content': [{'type': 'text', 'text': 'this is a text'}],
            },
            format='anthropic',
        ),
        lf.UserMessage('this is a text'),
    )
    self.assertEqual(
        lf.Message.from_value(
            {
                'role': 'assistant',
                'content': [{'type': 'text', 'text': 'this is a text'}],
            },
            format='anthropic',
        ),
        lf.AIMessage('this is a text'),
    )
    self.assertEqual(
        lf.Message.from_value(
            {
                'role': 'system',
                'content': [{'type': 'text', 'text': 'this is a text'}],
            },
            format='anthropic',
        ),
        lf.SystemMessage('this is a text'),
    )
    with self.assertRaisesRegex(ValueError, 'Unsupported role: .*'):
      lf.Message.from_value(
          {
              'role': 'function',
              'content': [{'type': 'text', 'text': 'this is a text'}],
          },
          format='anthropic',
      )

  def test_from_value_with_thoughts(self):
    message = lf.Message.from_anthropic_format(
        {
            'role': 'user',
            'content': [
                {
                    'type': 'thinking',
                    'thinking': 'this is a red round object',
                },
                {
                    'type': 'text',
                    'text': 'this is a apple',
                },
            ],
        },
    )
    self.assertEqual(message.text, 'this is a apple')
    self.assertEqual(message.thought, 'this is a red round object')

  def test_from_value_with_modalities(self):
    m = lf.Message.from_value(
        {
            'role': 'user',
            'content': [
                {
                    'type': 'text',
                    'text': 'What are the common words from'
                },
                {
                    'type': 'image',
                    'source': {
                        'type': 'base64',
                        'media_type': 'image/png',
                        'data': base64.b64encode(image_content).decode('utf-8'),
                    }
                },
                {
                    'type': 'text',
                    'text': 'and'
                },
                {
                    'type': 'document',
                    'source': {
                        'type': 'base64',
                        'media_type': 'application/pdf',
                        'data': base64.b64encode(pdf_content).decode('utf-8'),
                    }
                },
                {
                    'type': 'text',
                    'text': '?'
                },
            ],
        },
        format='anthropic',
    )
    self.assertEqual(
        m.text,
        'What are the common words from <<[[obj0]]>> and <<[[obj1]]>> ?'
    )
    self.assertIsInstance(m.obj0, lf_modalities.Image)
    self.assertEqual(m.obj0.mime_type, 'image/png')
    self.assertEqual(m.obj0.to_bytes(), image_content)

    self.assertIsInstance(m.obj1, lf_modalities.PDF)
    self.assertEqual(m.obj1.to_bytes(), pdf_content)


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for langfun.core.logging."""

import contextlib
import datetime
import inspect
import io
import unittest

from langfun.core import logging


class LoggingTest(unittest.TestCase):

  def test_use_log_level(self):
    self.assertEqual(logging.get_log_level(), 'info')
    with logging.use_log_level('debug'):
      self.assertEqual(logging.get_log_level(), 'debug')
      with logging.use_log_level(None):
        self.assertIsNone(logging.get_log_level(), None)
      self.assertEqual(logging.get_log_level(), 'debug')
    self.assertEqual(logging.get_log_level(), 'info')

  def test_log(self):
    entry = logging.log('info', 'hi', indent=1, x=1, y=2)
    self.assertEqual(entry.level, 'info')
    self.assertEqual(entry.message, 'hi')
    self.assertEqual(entry.indent, 1)
    self.assertEqual(entry.metadata, {'x': 1, 'y': 2})

    self.assertEqual(logging.debug('hi').level, 'debug')
    self.assertEqual(logging.info('hi').level, 'info')
    self.assertEqual(logging.warning('hi').level, 'warning')
    self.assertEqual(logging.error('hi').level, 'error')
    self.assertEqual(logging.fatal('hi').level, 'fatal')

  def test_log_without_system_logging(self):
    with contextlib.redirect_stdout(io.StringIO()) as stdout:
      logging.log('info', 'hi', indent=1, x=1, y=2, console=True)
    self.assertIn(
        'hi (metadata: {x=1, y=2})', stdout.getvalue()
    )

  def test_repr_html(self):
    def assert_color(entry, color):
      self.assertIn(f'background-color: {color}', entry._repr_html_())

    assert_color(logging.debug('hi', indent=0), '#EEEEEE')
    assert_color(logging.info('hi', indent=1), '#A3E4D7')
    assert_color(logging.warning('hi', x=1, y=2), '#F8C471')
    assert_color(logging.error('hi', indent=2, x=1, y=2), '#F5C6CB')
    assert_color(logging.fatal('hi', indent=2, x=1, y=2), '#F19CBB')

  def assert_html_content(self, html, expected):
    expected = inspect.cleandoc(expected).strip()
    actual = html.content.strip()
    if actual != expected:
      print(actual)
    self.assertEqual(actual, expected)

  def test_format(self):
    time = datetime.datetime(2024, 10, 10, 12, 30, 45)
    self.assertEqual(
        str(
            logging.LogEntry(
                level='info', message='hello\nworld',
                time=time, metadata=dict(x=1),
            )
        ),
        '12:30:45 INFO - hello\nworld (metadata: {x=1})',
    )
    self.assertIn(
        'LogEntry(',
        logging.LogEntry(
            level='info', message='hello\nworld',
            time=time, metadata=dict(x=1),
        ).format(text_format=False),
    )

  def test_html(self):
    time = datetime.datetime(2024, 10, 10, 12, 30, 45)
    self.assert_html_content(
        logging.LogEntry(
            level='info', message='5 + 2 > 3',
            time=time, metadata={}
        ).to_html(enable_summary_tooltip=False),
        """
        <details open class="pyglove log-entry log-info"><summary><div class="summary-title log-info"><span class="log-time">12:30:45</span><span class="log-summary">5 + 2 &gt; 3</span></div></summary><div class="complex_value"></div></details>
        """
    )
    self.assert_html_content(
        logging.LogEntry(
            level='error', message='This is a longer message: 5 + 2 > 3',
            time=time, metadata=dict(x=dict(z=1), y=2)
        ).to_html(
            extra_flags=dict(
                collapse_log_metadata_level=1,
            ),
            max_summary_len_for_str=10,
            enable_summary_tooltip=False,
        ),
        """
        <details open class="pyglove log-entry log-error"><summary><div class="summary-title log-error"><span class="log-time">12:30:45</span><span class="log-summary">This is a ...</span></div></summary><div class="complex_value"><span class="log-text">This is a longer message: 5 + 2 &gt; 3</span><div class="log-metadata"><details open class="pyglove dict"><summary><div class="summary-name">metadata<span class="tooltip">metadata</span></div><div class="summary-title">Dict(...)</div></summary><div class="complex-value dict"><details class="pyglove dict"><summary><div class="summary-name">x<span class="tooltip">metadata.x</span></div><div class="summary-title">Dict(...)</div></summary><div class="complex-value dict"><details open class="pyglove int"><summary><div class="summary-name">z<span class="tooltip">metadata.x.z</span></div><div class="summary-title">int</div></summary><span class="simple-value int">1</span></details></div></details><details open class="pyglove int"><summary><div class="summary-name">y<span class="tooltip">metadata.y</span></div><div class="summary-title">int</div></summary><span class="simple-value int">2</span></details></div></details></div></div></details>
        """
    )
    self.assert_html_content(
        logging.LogEntry(
            level='error', message='This is a longer message: 5 + 2 > 3',
            time=time, metadata=dict(x=dict(z=1), y=2)
        ).to_html(
            extra_flags=dict(
                max_summary_len_for_str=10,
            ),
            enable_summary_tooltip=False,
        ),
        """
        <details open class="pyglove log-entry log-error"><summary><div class="summary-title log-error"><span class="log-time">12:30:45</span><span class="log-summary">This is a longer message: 5 + 2 &gt; 3</span></div></summary><div class="complex_value"><div class="log-metadata"><details open class="pyglove dict"><summary><div class="summary-name">metadata<span class="tooltip">metadata</span></div><div class="summary-title">Dict(...)</div></summary><div class="complex-value dict"><details open class="pyglove dict"><summary><div class="summary-name">x<span class="tooltip">metadata.x</span></div><div class="summary-title">Dict(...)</div></summary><div class="complex-value dict"><details open class="pyglove int"><summary><div class="summary-name">z<span class="tooltip">metadata.x.z</span></div><div class="summary-title">int</div></summary><span class="simple-value int">1</span></details></div></details><details open class="pyglove int"><summary><div class="summary-name">y<span class="tooltip">metadata.y</span></div><div class="summary-title">int</div></summary><span class="simple-value int">2</span></details></div></details></div></div></details>
        """
    )

if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python code parsing."""

import inspect
import io
import re


_ID_REGEX = re.compile('^[a-zA-Z_\\-]*$')


def clean(code_text: str) -> str:
  """Cleans up Python code.

  LLM may generate code with markdown annotations, as well as minor syntax
  errors. This function removes such annotations and fixes minor syntax errors
  without extra LLM calls.

  Args:
    code_text: The code text to clean up.

  Returns:
    The cleaned up code text.
  """
  # TODO(daiyip): Deal with markdown in docstrings.
  code = io.StringIO()
  quote_char = None
  in_code = False
  i = 0
  in_comment = False
  while i < len(code_text):
    c = code_text[i]
    # Detect code block separator (```).
    if (not in_comment
        and quote_char is None
        and c == '`'
        and code_text[i:i + 3] == '```'):
      in_code = not in_code
      if in_code:
        i += 3
        continue
      else:
        break

    # Detect string literal boundary.
    if (in_code
        and not in_comment
        and c in ('\'', '"')
        and i > 0
        and code_text[i - 1] != '\\'):
      # Handle ''' and """.
      if code_text[i: i + 3] == c * 3:
        c = c * 3
        i += 2

      if quote_char is None:
        quote_char = c
      elif quote_char == c:
        # NOTE(daiyip): at times, LM forgets to escape quotes inside a string.
        # Thus we do some smart checking here to automatically correct such
        # case. This logic here is pretty involved in handling special cases.
        # We might want to revisit them later.

        # Peek forward to see if it could be a valid string.
        nt, nnt_start = _next_token(code_text, i + 1)
        if (len(c) == 3
            or nt in (',', '[', ']', '}', ')', '+', '*', '%', '\n', ':')):
          end_quote = True
        elif nt == ' ':
          # Detect if . could be a method invocation.
          # NOTE(daiyip): 'in' and 'not in' might have false positives. But
          # given the chance is low, we do not complicate the reasoning logic
          # for now.
          nnt, _ = _next_token(code_text, nnt_start, skip_whitespace=True)
          end_quote = nnt in ('+', '*', '%', '#', '[', 'in', 'not', ':')
        elif nt == '.':
          # Detect if . could be method invocation on string.
          nnt, nnnt_start = _next_token(code_text, nnt_start)
          nnnt, _ = _next_token(code_text, nnnt_start)
          end_quote = nnt.isidentifier() and nnnt == '('
        else:
          end_quote = False

        if end_quote:
          quote_char = None
        else:
          c = f'\\{c}'
    # Detect comment.
    elif c == '#' and quote_char is None:
      in_comment = True
    # Detect end-of-comment.
    elif c == '\n':
      # NOTE(daiyip): deal with cases that LM forgot to escape linebreaks
      # within strings.
      if quote_char is not None:
        # Only add \\ for ' and " (other than ''' and """).
        if len(quote_char) == 1:
          c = '\\n'
      else:
        in_comment = False

    if in_code:
      code.write(c)

    i += 1

  code = code.getvalue()
  if code:
    pos = code.find('\n')
    # Strip markdown code type. E.g. ```python
    if pos > 0 and _ID_REGEX.match(code[:pos]):
      code = code[pos:]
  else:
    # Maybe-code that resides not within a code markdown block.
    # Adding '\n' makes inspect.cleandoc to make right adjustment.
    code = '\n' + code_text
  return inspect.cleandoc(code).strip()


def _next_token(
    text: str,
    start: int = 0,
    skip_whitespace: bool = False
    ) -> tuple[str, int]:
  """Find the next token in a string with a start position."""
  token_start = start
  if skip_whitespace:
    while token_start < len(text) and text[token_start] in ' \t':
      token_start += 1
  token_end = token_start + 1
  if text[token_start].isalpha():
    while (token_end < len(text)
           and text[token_end].isalpha() or text[token_end] in '_'):
      token_end += 1
  return text[token_start:token_end], token_end



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for structured query."""

import asyncio
import inspect
import math
import time
from typing import Any
import unittest

import langfun.core as lf
from langfun.core import modalities
from langfun.core.llms import fake
from langfun.core.llms.cache import in_memory
from langfun.core.structured import mapping
from langfun.core.structured import querying
import pyglove as pg


class Activity(pg.Object):
  description: str


class Itinerary(pg.Object):
  day: pg.typing.Int[1, None]
  type: pg.typing.Enum['daytime', 'nighttime']
  activities: list[Activity]
  hotel: pg.typing.Str['.*Hotel'] | None


class QueryTest(unittest.TestCase):

  def assert_render(
      self,
      prompt,
      schema,
      examples: list[mapping.MappingExample] | None = None,
      system_message: str | None = None,
      *,
      expected_snippet: str,
      exact_match: bool = False,
      expected_modalities: int = 0,
      exepcted_system_message: str | None = None,
      **kwargs,
  ):
    m = querying.query(
        prompt, schema=schema, system_message=system_message, examples=examples,
        **kwargs, returns_message=True
    )
    self.assertIsNotNone(m.lm_input)
    if exact_match:
      self.assertEqual(expected_snippet, m.lm_input.text)
    else:
      self.assertIn(expected_snippet, m.lm_input.text)
    self.assertEqual(
        len([c for c in m.lm_input.chunk() if isinstance(c, lf.Modality)]),
        expected_modalities,
    )
    if system_message is not None:
      self.assertEqual(
          m.lm_input.system_message.text,
          exepcted_system_message,
      )

  def test_call(self):
    lm = fake.StaticSequence(['1'])
    self.assertEqual(querying.query('what is 1 + 0', int, lm=lm), 1)

    # Testing calling the same `lm` without copy.
    with self.assertRaises(IndexError):
      querying.query('what is 1 + 2', int, lm=lm)

    self.assertEqual(
        querying.query(
            'what is 1 + 0', int, lm=lm.clone(), returns_message=True
        ),
        lf.AIMessage(
            '1',
            result=1,
            score=1.0,
            logprobs=None,
            is_cached=False,
            usage=lf.LMSamplingUsage(364, 1, 365),
            tags=['lm-response', 'lm-output', 'transformed'],
        ),
    )
    self.assertEqual(
        querying.query(
            lf.Template('what is {{x}} + {{y}}', x=1, y=0), int, lm=lm.clone()
        ),
        1,
    )
    self.assertEqual(
        querying.query('what is {{x}} + {{y}}', int, x=1, y=0, lm=lm.clone()),
        1,
    )
    self.assertEqual(
        querying.query(
            'what is {{x}} + {{y}}',
            x=1,
            y=0,
            lm=fake.StaticResponse('The answer is one.'),
        ),
        'The answer is one.',
    )
    # Testing tool calls in the response.
    self.assertEqual(
        querying.query(
            'abc',
            Activity,
            lm=fake.StaticResponse(
                lf.AIMessage(
                    'Here is the answer.',
                    tool_calls=[lf.AIMessage('Activity(description="hello")')],
                    ),
            ),
        ),
        Activity(description='hello'),
    )
    # Test completing a partial object.
    self.assertEqual(
        querying.query(
            Activity.partial(),
            lm=fake.StaticResponse('Activity(description="hello")'),
        ),
        Activity(description='hello'),
    )

  def test_render_with_system_message(self):
    lm = fake.StaticResponse('1')
    self.assert_render(
        'What is {{x}} + {{y}}?',
        schema=None,
        system_message='You are a helpful assistant.',
        x=1,
        y=2,
        lm=lm.clone(),
        expected_snippet='What is 1 + 2?',
        exepcted_system_message='You are a helpful assistant.',
    )

  def test_str_to_structure_render(self):
    lm = fake.StaticResponse('1')
    self.assert_render(
        'What is {{x}} + {{y}}?',
        int,
        x=1,
        y=2,
        lm=lm.clone(),
        expected_snippet=(
            'Please respond to the last REQUEST with OUTPUT PYTHON OBJECT '
            'only according to OUTPUT PYTHON TYPE.\n\n'
            'REQUEST:\n  1 + 1 =\n\n'
            'OUTPUT PYTHON TYPE:\n'
            '  Answer\n\n'
            '  ```python\n'
            '  class Answer:\n'
            '    final_answer: int\n'
            '  ```\n\n'
            'OUTPUT PYTHON OBJECT:\n'
            '  ```python\n'
            '  output = Answer(\n'
            '    final_answer=2\n'
            '  )\n'
            '  ```\n\n'
            'REQUEST:\n'
            '  What is 1 + 2?\n\n'
            'OUTPUT PYTHON TYPE:\n'
            '  int\n\n'
            'OUTPUT PYTHON OBJECT:'
        ),
    )

  def test_str_to_structure_render_custom_template(self):
    lm = fake.StaticResponse('1')
    self.assert_render(
        'What is {{x}} + {{y}}?',
        int,
        x=1,
        y=2,
        lm=lm.clone(),
        template_str='!!{{ DEFAULT }}!!',
        expected_snippet=(
            '!!Please respond to the last REQUEST with OUTPUT PYTHON OBJECT '
            'only according to OUTPUT PYTHON TYPE.\n\n'
            'REQUEST:\n  1 + 1 =\n\n'
            'OUTPUT PYTHON TYPE:\n'
            '  Answer\n\n'
            '  ```python\n'
            '  class Answer:\n'
            '    final_answer: int\n'
            '  ```\n\n'
            'OUTPUT PYTHON OBJECT:\n'
            '  ```python\n'
            '  output = Answer(\n'
            '    final_answer=2\n'
            '  )\n'
            '  ```\n\n'
            'REQUEST:\n'
            '  What is 1 + 2?\n\n'
            'OUTPUT PYTHON TYPE:\n'
            '  int\n\n'
            'OUTPUT PYTHON OBJECT:!!'
        ),
    )

  def test_str_to_str_render(self):
    lm = fake.StaticResponse('1')
    self.assert_render(
        'What is {{x}} + {{y}}?',
        None,
        x=1,
        y=2,
        lm=lm.clone(),
        expected_snippet='What is 1 + 2?',
        exact_match=True,
    )

  def test_structure_to_structure_render(self):
    lm = fake.StaticResponse('[1]')
    self.assert_render(
        [1],
        list[int],
        x=1,
        y=2,
        lm=lm.clone(),
        expected_snippet=(
            '\n\nREQUEST:\n  ```python\n  [\n    1\n  ]\n  ```\n\n'
        ),
    )

  def test_structure_to_str_render(self):
    lm = fake.StaticResponse('[1]')
    self.assert_render(
        [1], None, x=1, y=2, lm=lm, expected_snippet='`[1]`', exact_match=True
    )

  def test_root_modality_to_structure_render(self):
    lm = fake.StaticResponse('1')
    self.assert_render(
        modalities.Image.from_bytes(b'mock_image'),
        int,
        lm=lm,
        expected_snippet='\n\nREQUEST:\n  <<[[input]]>>\n\n',
        expected_modalities=1,
    )

  def test_root_modality_to_str_render(self):
    lm = fake.StaticResponse('1')
    self.assert_render(
        modalities.Image.from_bytes(b'mock_image'),
        None,
        lm=lm,
        expected_snippet='<<[[input]]>>',
        exact_match=True,
        expected_modalities=1,
    )

  def test_str_with_modality_to_str_render(self):
    lm = fake.StaticResponse('A cat and a mouse.')
    self.assert_render(
        'What are these? {{this_image}} and {{that_image}}',
        None,
        this_image=modalities.Image.from_bytes(b'cat_image'),
        that_image=modalities.Image.from_bytes(b'mouse_image'),
        lm=lm,
        expected_snippet=(
            'What are these? <<[[this_image]]>> and <<[[that_image]]>>'
        ),
        exact_match=True,
        expected_modalities=2,
    )

  def test_structure_with_modality_to_str_render(self):
    lm = fake.StaticResponse('A cat and a mouse.')
    self.assert_render(
        [
            modalities.Image.from_bytes(b'cat_image'),
            modalities.Image.from_bytes(b'mouse_image'),
        ],
        None,
        lm=lm,
        expected_snippet='`[<<[[input[0]]]>>, <<[[input[1]]]>>]`',
        exact_match=True,
        expected_modalities=2,
    )

  def test_structure_with_modality_to_structure_render(self):
    lm = fake.StaticResponse('["cat", "mouse"]')
    self.assert_render(
        [
            modalities.Image.from_bytes(b'cat_image'),
            modalities.Image.from_bytes(b'mouse_image'),
        ],
        list[str],
        lm=lm,
        expected_snippet=inspect.cleandoc("""
            REQUEST:
              ```python
              [
                <<[[input[0]]]>>,
                <<[[input[1]]]>>
              ]
              ```
            """),
        expected_modalities=2,
    )

  def test_structure_with_modality_and_examples_to_structure_render(self):
    lm = fake.StaticResponse('["cat", "mouse"]')
    self.assert_render(
        [
            modalities.Image.from_bytes(b'cat_image'),
            modalities.Image.from_bytes(b'mouse_image'),
        ],
        list[str],
        examples=[
            mapping.MappingExample(
                input=[modalities.Image.from_bytes(b'dog_image')],
                schema=list[str],
                output=['dog'],
            ),
        ],
        lm=lm,
        expected_snippet=inspect.cleandoc("""
            REQUEST:
              ```python
              [
                <<[[examples[0].input[0]]]>>
              ]
              ```

            OUTPUT PYTHON TYPE:
              list[str]

            OUTPUT PYTHON OBJECT:
              ```python
              output = [
                'dog'
              ]
              ```


            REQUEST:
              ```python
              [
                <<[[input[0]]]>>,
                <<[[input[1]]]>>
              ]
              ```

            OUTPUT PYTHON TYPE:
              list[str]

            OUTPUT PYTHON OBJECT:
            """),
        expected_modalities=3,
    )

  def test_multiple_queries(self):
    self.assertEqual(
        querying.query(
            'Compute 1 + 2',
            int,
            lm=[
                fake.StaticResponse('1'),
                fake.StaticResponse('2'),
            ],
            num_samples=[1, 2],
        ),
        [1, 2, 2]
    )
    self.assertEqual(
        querying.query(
            'Compute 1 + 2',
            int,
            lm=[
                fake.StaticResponse('1'),
                fake.StaticResponse('2'),
            ],
            num_samples=2,
        ),
        [1, 1, 2, 2]
    )
    self.assertEqual(
        querying.query(
            'Compute 1 + 2',
            int,
            lm=[
                fake.StaticResponse('1'),
                fake.StaticResponse('abc'),
            ],
            num_samples=[1, 2],
        ),
        [1]
    )
    self.assertEqual(
        querying.query(
            'Compute 1 + 2',
            int,
            default=0,
            lm=[
                fake.StaticResponse('1'),
                fake.StaticResponse('abc'),
            ],
            num_samples=[1, 2],
        ),
        [1, 0, 0]
    )
    results = querying.query(
        'Compute 1 + 2',
        int,
        default=0,
        lm=[
            fake.StaticResponse('1'),
            fake.StaticResponse('abc'),
        ],
        returns_message=True,
    )
    self.assertEqual([r.text for r in results], ['1', 'abc'])
    self.assertEqual([r.result for r in results], [1, 0])

  def test_from_protocol(self):
    self.assertIs(
        querying.LfQuery.from_protocol('python'), querying._LfQueryPythonV2
    )
    self.assertIs(
        querying.LfQuery.from_protocol('python:1.0'),
        querying._LfQueryPythonV1
    )

    class MyLfQuery(querying.LfQuery):
      protocol = 'yaml'
      version = '1.0'

    self.assertIs(
        querying.LfQuery.from_protocol('yaml:1.0'),
        MyLfQuery
    )
    self.assertIs(
        querying.LfQuery.from_protocol('yaml'),
        MyLfQuery
    )

    with self.assertRaisesRegex(
        ValueError, 'Version .* is already registered'
    ):
      class MyLfQuery2(querying.LfQuery):  # pylint: disable=unused-variable
        protocol = 'yaml'
        version = '1.0'

    with self.assertRaisesRegex(
        ValueError, 'Version \'2.0\' is not supported for protocol \'yaml\''
    ):
      querying.LfQuery.from_protocol('yaml:2.0')

    class MyLfQuery3(querying.LfQuery):  # pylint: disable=unused-variable
      protocol = 'yaml'
      version = '3.0'

    with self.assertRaisesRegex(
        ValueError, 'Multiple versions found for protocol \'yaml\''
    ):
      querying.LfQuery.from_protocol('yaml')

    with self.assertRaisesRegex(
        ValueError, 'Protocol \'text\' is not supported'
    ):
      querying.LfQuery.from_protocol('text')

    with self.assertRaisesRegex(
        ValueError, 'Protocol \'text\' is not supported'
    ):
      querying.LfQuery.from_protocol('text:1.0')

  def test_query_prompt(self):
    self.assertEqual(
        querying.query_prompt('what is this?', int),
        inspect.cleandoc("""
            Please respond to the last REQUEST with OUTPUT PYTHON OBJECT only according to OUTPUT PYTHON TYPE.

            REQUEST:
              1 + 1 =

            OUTPUT PYTHON TYPE:
              Answer

              ```python
              class Answer:
                final_answer: int
              ```

            OUTPUT PYTHON OBJECT:
              ```python
              output = Answer(
                final_answer=2
              )
              ```

            REQUEST:
              what is this?

            OUTPUT PYTHON TYPE:
              int

            OUTPUT PYTHON OBJECT:
            """),
    )

  def test_query_prompt_with_metadata(self):
    self.assertIn(
        'x',
        querying.query_prompt(
            'what is this?',
            metadata_x=1
        ).metadata
    )
    self.assertIn(
        'x',
        querying.query_prompt(
            'what is this?',
            int,
            metadata_x=1
        ).metadata
    )

  def test_query_prompt_with_unrooted_template(self):
    output = querying.query_prompt(
        pg.Dict(
            input=lf.Template(
                'what is {{image}}',
                image=modalities.Image.from_bytes(b'mock_image')
            )
        ).input,
    )
    self.assertIsNotNone(output.get_modality('image'))

  def test_query_and_reduce(self):
    self.assertEqual(
        querying.query_and_reduce(
            'Compute 1 + 1',
            int,
            reduce=sum,
            lm=[
                fake.StaticResponse('1'),
                fake.StaticResponse('2'),
            ],
            num_samples=[1, 2],
        ),
        5
    )
    self.assertEqual(
        querying.query_and_reduce(
            'Compute 1 + 1',
            int,
            reduce=sum,
            lm=fake.StaticResponse('2'),
        ),
        2
    )

  def test_query_output(self):
    self.assertEqual(
        querying.query_output(
            lf.AIMessage('1'),
            int,
        ),
        1,
    )

  def test_query_reward(self):

    class Answer(pg.Object):
      final_answer: int

      def __reward__(self, inputs: lf.Template) -> None:
        diff = abs(self.final_answer - (inputs.x + inputs.y))
        # Center screwed sigmoid scaled to [-1.0 and 1.0].
        return 4 / (1 + math.exp(diff)) - 1.0

    # Case 1: Reward function based on input and output.
    self.assertEqual(
        querying.query_reward(
            mapping.MappingExample(
                input=lf.Template('{{x}} + {{y}}', x=1, y=1),
                schema=Answer,
                output=Answer(final_answer=2),
            ),
            'Answer(2)'
        ),
        1.0
    )
    self.assertEqual(
        querying.query_reward(
            mapping.MappingExample(
                input=lf.Template('{{x}} + {{y}}', x=2, y=3),
                output=Answer(final_answer=2),
            ).to_json_str(),
            'Answer(5)'
        ),
        1.0
    )

    # Case 2: Reward function based on input, result and expected output.
    class Answer2(pg.Object):
      final_answer: int

      def __reward__(self, inputs: lf.Template, expected_output: 'Answer2'):
        return (
            1.0 if self.final_answer == expected_output.final_answer else -1.0
        )

    self.assertEqual(
        querying.query_reward(
            mapping.MappingExample(
                input=lf.Template('{{x}} + {{y}}', x=1, y=1),
                output=Answer2(final_answer=2),
            ),
            'Answer2(3)'
        ),
        -1.0
    )

    # Case 3: Reward function based on input, result, expected output
    # and metadata.
    class Answer3(pg.Object):
      final_answer: int

      def __reward__(self,
                     inputs: lf.Template,
                     expected_output: 'Answer3',
                     metadata: dict[str, Any]):
        del inputs
        return (
            1.0 if self.final_answer == expected_output.final_answer else -1.0
        ) * metadata['weight']

    self.assertEqual(
        querying.query_reward(
            mapping.MappingExample(
                input=lf.Template('{{x}} + {{y}}', x=1, y=1),
                output=Answer3(final_answer=2),
                metadata=dict(weight=0.5)
            ),
            'Answer3(3)'
        ),
        -0.5
    )

    # Case 4: No reward function is provided.
    class Answer4(pg.Object):
      final_answer: int

    self.assertIsNone(
        querying.query_reward(
            mapping.MappingExample(
                input=lf.Template('{{x}} + {{y}}', x=1, y=1),
                output=Answer4(final_answer=2),
            ),
            'Answer2(2)'
        )
    )

    # Case 5: Not a structured output.
    self.assertIsNone(
        querying.query_reward(
            mapping.MappingExample(
                input=lf.Template('{{x}} + {{y}}', x=1, y=1),
                output='2',
            ),
            '2'
        )
    )

    # Case 6: Bad reward function.
    class Answer5(pg.Object):
      final_answer: int

      def __reward__(self):
        return 0.0

    with self.assertRaisesRegex(
        TypeError, '.*Answer5.__reward__` should have signature'
    ):
      querying.query_reward(
          mapping.MappingExample(
              input=lf.Template('{{x}} + {{y}}', x=1, y=1),
              output=Answer5(final_answer=2),
          ),
          'Answer5(2)'
      )


class LfQueryPythonV1Test(unittest.TestCase):

  def test_render_no_examples(self):
    l = querying.LfQuery.from_protocol('python:1.0')(
        input=lf.AIMessage('Compute 12 / 6 + 2.'), schema=int
    )
    self.assertEqual(
        l.render().text,
        inspect.cleandoc("""
            Please respond to the last INPUT_OBJECT with OUTPUT_OBJECT according to OUTPUT_TYPE.

            INPUT_OBJECT:
              1 + 1 =

            OUTPUT_TYPE:
              Answer

              ```python
              class Answer:
                final_answer: int
              ```

            OUTPUT_OBJECT:
              ```python
              Answer(
                final_answer=2
              )
              ```

            INPUT_OBJECT:
              Compute 12 / 6 + 2.

            OUTPUT_TYPE:
              int

            OUTPUT_OBJECT:
            """),
    )

  def test_render(self):
    l = querying.LfQuery.from_protocol('python:1.0')(
        input=lf.AIMessage('Compute 12 / 6 + 2.'),
        schema=int,
        examples=[
            mapping.MappingExample(
                input='What is the answer of 1 plus 1?', output=2
            ),
            mapping.MappingExample(
                input='Compute the value of 3 + (2 * 6).', output=15
            ),
        ],
    )
    self.assertEqual(
        l.render().text,
        inspect.cleandoc("""
            Please respond to the last INPUT_OBJECT with OUTPUT_OBJECT according to OUTPUT_TYPE.

            INPUT_OBJECT:
              1 + 1 =

            OUTPUT_TYPE:
              Answer

              ```python
              class Answer:
                final_answer: int
              ```

            OUTPUT_OBJECT:
              ```python
              Answer(
                final_answer=2
              )
              ```

            INPUT_OBJECT:
              What is the answer of 1 plus 1?

            OUTPUT_TYPE:
              int

            OUTPUT_OBJECT:
              ```python
              2
              ```

            INPUT_OBJECT:
              Compute the value of 3 + (2 * 6).

            OUTPUT_TYPE:
              int

            OUTPUT_OBJECT:
              ```python
              15
              ```


            INPUT_OBJECT:
              Compute 12 / 6 + 2.

            OUTPUT_TYPE:
              int

            OUTPUT_OBJECT:
            """),
    )

  def test_invocation(self):
    lm_input = lf.UserMessage('3-day itineraries to San Francisco')
    parse_structured_response = inspect.cleandoc(
        """
        ```python
        [
            Itinerary(
                day=1,
                type='daytime',
                activities=[
                    Activity(description='Arrive in San Francisco and check into your hotel.'),
                    Activity(description='Take a walk around Fisherman\\'s Wharf and have dinner at one of the many seafood restaurants.'),
                    Activity(description='Visit Pier 39 and see the sea lions.'),
                ], 
                hotel=None),
            Itinerary(
                day=2,
                type='daytime',
                activities=[
                    Activity(description='Take a ferry to Alcatraz Island and tour the infamous prison.'),
                    Activity(description='Take a walk across the Golden Gate Bridge.'),
                    Activity(description='Visit the Japanese Tea Garden in Golden Gate Park.'),
                ], 
                hotel=None),
            Itinerary(
                day=3,
                type='daytime',
                activities=[
                    Activity(description='Visit the de Young Museum and see the collection of American art.'),
                    Activity(description='Visit the San Francisco Museum of Modern Art.'),
                    Activity(description='Take a cable car ride.'),
                ], 
                hotel=None),
        ]
        ```
        """)
    with lf.context(
        lm=fake.StaticSequence(
            [parse_structured_response],
        ),
        override_attrs=True,
    ):
      l = querying.LfQuery.from_protocol('python:1.0')(
          input=lm_input,
          schema=[Itinerary],
          examples=[
              mapping.MappingExample(
                  input=inspect.cleandoc("""
                      Find the alternatives of expressing \"feeling great\".
                      """),
                  schema={'expression': str, 'words': list[str]},
                  output={
                      'expression': 'feeling great',
                      'words': [
                          'Ecstatic',
                          'Delighted',
                          'Wonderful',
                          'Enjoyable',
                          'Fantastic',
                      ],
                  },
              )
          ],
      )
      r = l()
      self.assertEqual(len(r.result), 3)
      self.assertIsInstance(r.result[0], Itinerary)
      self.assertEqual(len(r.result[0].activities), 3)
      self.assertIsNone(r.result[0].hotel)


class LfQueryPythonV2Test(unittest.TestCase):

  def test_render_no_examples(self):
    l = querying.LfQuery.from_protocol('python:2.0')(
        input=lf.AIMessage('Compute 12 / 6 + 2.'), schema=int
    )
    self.assertEqual(
        l.render().text,
        inspect.cleandoc("""
            Please respond to the last REQUEST with OUTPUT PYTHON OBJECT only according to OUTPUT PYTHON TYPE.

            REQUEST:
              1 + 1 =

            OUTPUT PYTHON TYPE:
              Answer

              ```python
              class Answer:
                final_answer: int
              ```

            OUTPUT PYTHON OBJECT:
              ```python
              output = Answer(
                final_answer=2
              )
              ```

            REQUEST:
              Compute 12 / 6 + 2.

            OUTPUT PYTHON TYPE:
              int

            OUTPUT PYTHON OBJECT:
            """),
    )

  def test_render_with_examples(self):
    l = querying.LfQuery.from_protocol('python:2.0')(
        input=lf.AIMessage('Compute 12 / 6 + 2.'),
        schema=int,
        examples=[
            mapping.MappingExample(
                input='What is the answer of 1 plus 1?', output=2
            ),
            mapping.MappingExample(
                input='Compute the value of 3 + (2 * 6).', output=15
            ),
        ],
    )
    self.assertEqual(
        l.render().text,
        inspect.cleandoc("""
            Please respond to the last REQUEST with OUTPUT PYTHON OBJECT only according to OUTPUT PYTHON TYPE.

            REQUEST:
              1 + 1 =

            OUTPUT PYTHON TYPE:
              Answer

              ```python
              class Answer:
                final_answer: int
              ```

            OUTPUT PYTHON OBJECT:
              ```python
              output = Answer(
                final_answer=2
              )
              ```

            REQUEST:
              What is the answer of 1 plus 1?

            OUTPUT PYTHON TYPE:
              int

            OUTPUT PYTHON OBJECT:
              ```python
              output = 2
              ```

            REQUEST:
              Compute the value of 3 + (2 * 6).

            OUTPUT PYTHON TYPE:
              int

            OUTPUT PYTHON OBJECT:
              ```python
              output = 15
              ```


            REQUEST:
              Compute 12 / 6 + 2.

            OUTPUT PYTHON TYPE:
              int

            OUTPUT PYTHON OBJECT:
            """),
    )

  def test_bad_response(self):
    with self.assertRaisesRegex(
        mapping.MappingError,
        'name .* is not defined',
    ):
      querying.query('Compute 1 + 2', int, lm=fake.StaticSequence(['a2']))

  def test_not_allowed_code(self):
    lm = fake.StaticResponse(
        """
        class Foo:
          pass
        """
    )
    with self.assertRaisesRegex(
        mapping.MappingError,
        'Class definition is not allowed',
    ):
      querying.query('Compute 1 + 2', int, lm=lm)

  def test_autofix(self):
    lm = fake.StaticSequence([
        '=1',
        inspect.cleandoc("""
            CorrectedCode(
                corrected_code='1',
            )
            """),
    ])
    self.assertEqual(querying.query('what is 1 + 0', int, lm=lm, autofix=3), 1)

  def test_response_postprocess(self):
    self.assertEqual(
        querying.query(
            'Compute 1 + 2',
            lm=fake.StaticResponse('<!-- some comment-->\n3'),
            response_postprocess=lambda x: x.split('\n')[1]),
        '3'
    )
    self.assertEqual(
        querying.query(
            'Compute 1 + 2', int,
            lm=fake.StaticResponse('<!-- some comment-->\n3'),
            response_postprocess=lambda x: x.split('\n')[1]),
        3
    )

  def test_render(self):
    l = querying.LfQuery.from_protocol('python:2.0')(
        input=lf.AIMessage('Compute 12 / 6 + 2.'),
        schema=int,
        examples=[
            mapping.MappingExample(
                input='What is the answer of 1 plus 1?', output=2
            ),
            mapping.MappingExample(
                input='Compute the value of 3 + (2 * 6).', output=15
            ),
        ],
    )
    self.assertEqual(
        l.render().text,
        inspect.cleandoc("""
            Please respond to the last REQUEST with OUTPUT PYTHON OBJECT only according to OUTPUT PYTHON TYPE.

            REQUEST:
              1 + 1 =

            OUTPUT PYTHON TYPE:
              Answer

              ```python
              class Answer:
                final_answer: int
              ```

            OUTPUT PYTHON OBJECT:
              ```python
              output = Answer(
                final_answer=2
              )
              ```

            REQUEST:
              What is the answer of 1 plus 1?

            OUTPUT PYTHON TYPE:
              int

            OUTPUT PYTHON OBJECT:
              ```python
              output = 2
              ```

            REQUEST:
              Compute the value of 3 + (2 * 6).

            OUTPUT PYTHON TYPE:
              int

            OUTPUT PYTHON OBJECT:
              ```python
              output = 15
              ```


            REQUEST:
              Compute 12 / 6 + 2.

            OUTPUT PYTHON TYPE:
              int

            OUTPUT PYTHON OBJECT:
            """),
    )


class LfQueryJsonV1Test(unittest.TestCase):

  def test_render_no_examples(self):
    l = querying.LfQuery.from_protocol('json:1.0')(
        input=lf.AIMessage('Compute 12 / 6 + 2.'), schema=int
    )
    self.assertEqual(
        l.render().text,
        inspect.cleandoc("""
            Please respond to the last INPUT_OBJECT with JSON according to SCHEMA:

            INSTRUCTIONS:
              1. If the schema has `_type`, carry it over to the JSON output.
              2. If a field from the schema cannot be extracted from the response, use null as the JSON value.

            INPUT_OBJECT:
              1 + 1 =

            SCHEMA:
              {"result": {"_type": "langfun.core.structured.query.Answer", "final_answer": int}}

            JSON:
              {"result": {"_type": "langfun.core.structured.query.Answer", "final_answer": 2}}

            INPUT_OBJECT:
              Compute 12 / 6 + 2.

            SCHEMA:
              {"result": int}

            JSON:
            """),
    )

  def test_render(self):
    l = querying.LfQuery.from_protocol('json:1.0')(
        input=lf.AIMessage('Compute 12 / 6 + 2.'),
        schema=int,
        examples=[
            mapping.MappingExample('What is the answer of 1 plus 1?', 2),
            mapping.MappingExample('Compute the value of 3 + (2 * 6).', 15),
        ],
    )
    self.assertEqual(
        l.render().text,
        inspect.cleandoc("""
            Please respond to the last INPUT_OBJECT with JSON according to SCHEMA:

            INSTRUCTIONS:
              1. If the schema has `_type`, carry it over to the JSON output.
              2. If a field from the schema cannot be extracted from the response, use null as the JSON value.

            INPUT_OBJECT:
              1 + 1 =

            SCHEMA:
              {"result": {"_type": "langfun.core.structured.query.Answer", "final_answer": int}}

            JSON:
              {"result": {"_type": "langfun.core.structured.query.Answer", "final_answer": 2}}

            INPUT_OBJECT:
              What is the answer of 1 plus 1?

            SCHEMA:
              {"result": int}

            JSON:
              {"result": 2}

            INPUT_OBJECT:
              Compute the value of 3 + (2 * 6).

            SCHEMA:
              {"result": int}

            JSON:
              {"result": 15}


            INPUT_OBJECT:
              Compute 12 / 6 + 2.

            SCHEMA:
              {"result": int}

            JSON:
            """),
    )

  def test_invocation(self):
    lm_input = lf.UserMessage('3-day itineraries to San Francisco')
    parse_structured_response = (
        lf.LangFunc(
            """
        {"result": [
          {
            "_type": {{itinerary_type}},
            "day": 1,
            "type": "daytime",
            "activities": [
              {
                "_type": {{activity_type}},
                "description": "Arrive in San Francisco and check into your hotel."
              },
              {
                "_type": {{activity_type}},
                "description": "Take a walk around Fisherman's Wharf and have dinner at one of the many seafood restaurants."
              },
              {
                "_type": {{activity_type}},
                "description": "Visit Pier 39 and see the sea lions."
              }
            ],
            "hotel": null
          },
          {
              "_type": {{itinerary_type}},
              "day": 2,
              "type": "daytime",
              "activities": [
                {
                  "_type": {{activity_type}},
                  "description": "Take a ferry to Alcatraz Island and tour the infamous prison."
                },
                {
                  "_type": {{activity_type}},
                  "description": "Take a walk across the Golden Gate Bridge."
                },
                {
                  "_type": {{activity_type}},
                  "description": "Visit the Japanese Tea Garden in Golden Gate Park."
                }
              ], 
              "hotel": null
           },
           {
              "_type": {{itinerary_type}},
              "day": 3,
              "type": "daytime",
              "activities": [
                {
                  "_type": {{activity_type}},
                  "description": "Visit the de Young Museum and see the collection of American art."
                },
                {
                  "_type": {{activity_type}},
                  "description": "Visit the San Francisco Museum of Modern Art."
                },
                {
                  "_type": {{activity_type}},
                  "description": "Take a cable car ride."
                }
              ],
              "hotel": null
            }
          ]}
        """,
            itinerary_type=f'"{Itinerary.__type_name__}"',
            activity_type=f'"{Activity.__type_name__}"',
        )
        .render()
        .text
    )
    with lf.context(
        lm=fake.StaticSequence(
            [parse_structured_response],
        ),
        override_attrs=True,
    ):
      l = querying.LfQuery.from_protocol('json:1.0')(
          input=lm_input,
          schema=[Itinerary],
          examples=[
              mapping.MappingExample(
                  input=inspect.cleandoc("""
                      Find the alternatives of expressing \"feeling great\".
                      """),
                  schema={'expression': str, 'words': list[str]},
                  output={
                      'expression': 'feeling great',
                      'words': [
                          'Ecstatic',
                          'Delighted',
                          'Wonderful',
                          'Enjoyable',
                          'Fantastic',
                      ],
                  },
              )
          ],
      )
      r = l()
      self.assertEqual(len(r.result), 3)
      self.assertIsInstance(r.result[0], Itinerary)
      self.assertEqual(len(r.result[0].activities), 3)
      self.assertIsNone(r.result[0].hotel)

  def test_bad_transform(self):
    with in_memory.lm_cache() as cache:
      with self.assertRaisesRegex(
          mapping.MappingError,
          'No JSON dict in the output',
      ):
        querying.query(
            'Compute 1 + 2', int, 
            lm=fake.StaticSequence(['3']),
            protocol='json', cache_seed=1
        )
      # Make sure bad mapping does not impact cache.
      self.assertEqual(len(cache), 0)

  def test_query(self):
    lm = fake.StaticResponse('{"result": 1}')
    self.assertEqual(
        querying.query('what is 1 + 0', int, lm=lm, protocol='json'), 1
    )
    self.assertEqual(
        querying.query('what is 1 + 0', int, lm=lm, protocol='json:1.0'), 1
    )
    with querying.query_protocol('json'):
      self.assertEqual(
          querying.query('what is 1 + 0', int, lm=lm), 1
      )

  def test_aquery(self):
    with lf.context(lm=fake.StaticResponse('{"result": 1}')):
      self.assertEqual(
          asyncio.run(
              querying.aquery('what is 1 + 0', int, protocol='json')
          ),
          1
      )


class QueryInvocationTest(unittest.TestCase):

  def test_basics(self):
    lm = fake.StaticSequence([
        'Activity(description="hi"',
    ])
    with querying.track_queries() as queries:
      querying.query('foo', Activity, default=None, lm=lm, invocation_id='123')

    self.assertEqual(queries[0].id, '123')
    self.assertTrue(queries[0].has_error)
    self.assertIsInstance(queries[0].error, pg.ErrorInfo)
    self.assertIn('MappingError', queries[0].error.tag)
    self.assertIsNone(queries[0].output)

  def test_kwargs(self):
    lm = fake.StaticSequence([
        'Activity(description="hi")',
    ])
    with querying.track_queries() as queries:
      querying.query(
          'foo {{x}}',
          Activity,
          lm=lm,
          system_message='system message',
          x=1,
      )
    self.assertTrue(queries[0].protocol.startswith('python:'))
    self.assertEqual(
        list(queries[0].kwargs.keys()),
        ['x', 'metadata_system_message']
    )
    self.assertIn('foo 1', queries[0].lm_request.text)
    self.assertEqual(queries[0].lm_request.system_message, 'system message')

  def test_to_html(self):
    lm = fake.StaticSequence([
        'Activity(description="hi")',
    ])
    with querying.track_queries() as queries:
      querying.query('foo', Activity, lm=lm)

    self.assertIn('query@', queries[0].id)
    self.assertIn('schema', queries[0].to_html_str())
    self.assertEqual(queries[0].lm_response.score, 1.0)
    self.assertFalse(queries[0].lm_response.is_cached)

  def test_as_mapping_example(self):
    lm = fake.StaticSequence([
        'Activity(description="hi")',
    ])
    with querying.track_queries() as queries:
      querying.query(lf.Template('foo {{x}}', x=1), Activity, lm=lm)
    mapping_example = queries[0].as_mapping_example()
    self.assertTrue(pg.eq(mapping_example.input, lf.Template('foo {{x}}', x=1)))
    self.assertEqual(mapping_example.schema.spec.cls, Activity)
    self.assertEqual(mapping_example.output, Activity(description='hi'))

    lm = fake.StaticSequence(['Activity(description="hi"'])
    with querying.track_queries() as queries:
      querying.query(
          lf.Template('foo {{x}}', x=1), Activity, default=None, lm=lm
      )

    self.assertTrue(queries[0].has_error)
    mapping_example = queries[0].as_mapping_example()
    self.assertTrue(pg.eq(mapping_example.input, lf.Template('foo {{x}}', x=1)))
    self.assertEqual(mapping_example.schema.spec.cls, Activity)
    self.assertEqual(mapping_example.output, 'Activity(description="hi"')


class TrackQueriesTest(unittest.TestCase):

  def test_query_without_schema(self):
    lm = fake.StaticSequence([
        'bar',
    ])
    with querying.track_queries() as queries:
      querying.query('foo', lm=lm)
    self.assertEqual(len(queries), 1)
    self.assertEqual(queries[0].lm_response, 'bar')
    self.assertTrue(pg.eq(queries[0].input, lf.Template('foo')))
    self.assertEqual(queries[0].lm_request, 'foo')
    self.assertIsNone(queries[0].schema)
    self.assertEqual(queries[0].protocol, 'python')
    self.assertEqual(queries[0].default, lf.RAISE_IF_HAS_ERROR)

    self.assertEqual(queries[0].output, 'bar')
    self.assertFalse(queries[0].has_error)
    self.assertIsNone(queries[0].error)
    self.assertIsNotNone(queries[0].end_time)
    self.assertIsNotNone(queries[0].usage_summary)

  def test_query_with_schema(self):
    lm = fake.StaticSequence([
        'Activity(description="hi")',
    ])
    with querying.track_queries() as queries:
      querying.query('foo', Activity, lm=lm)
    self.assertEqual(len(queries), 1)
    self.assertTrue(pg.eq(queries[0].input, lf.Template('foo')))
    self.assertIn('foo', queries[0].lm_request.text)
    self.assertIn('Please respond to ', queries[0].lm_request.text)
    self.assertEqual(queries[0].schema.spec.cls, Activity)
    self.assertEqual(queries[0].default, lf.RAISE_IF_HAS_ERROR)
    self.assertTrue(queries[0].protocol.startswith('python:'))

    self.assertTrue(pg.eq(queries[0].output, Activity(description='hi')))
    self.assertFalse(queries[0].has_error)
    self.assertIsNone(queries[0].error)
    self.assertIsNotNone(queries[0].end_time)
    self.assertIsNotNone(queries[0].usage_summary)

  def test_query_with_schema_and_default(self):
    lm = fake.StaticSequence([
        'Activity(description="hi")',
    ])
    with querying.track_queries() as queries:
      querying.query('foo', Activity, default=None, lm=lm)
    self.assertEqual(len(queries), 1)
    self.assertTrue(pg.eq(queries[0].input, lf.Template('foo')))
    self.assertIn('foo', queries[0].lm_request.text)
    self.assertIn('Please respond to ', queries[0].lm_request.text)
    self.assertEqual(queries[0].schema.spec.cls, Activity)
    self.assertIsNone(queries[0].default)
    self.assertTrue(queries[0].protocol.startswith('python:'))

    self.assertTrue(pg.eq(queries[0].output, Activity(description='hi')))
    self.assertFalse(queries[0].has_error)
    self.assertIsNone(queries[0].error)
    self.assertIsNotNone(queries[0].end_time)
    self.assertIsNotNone(queries[0].usage_summary)

  def test_oop_error_for_query_with_schema(self):
    lm = fake.StaticSequence([
        'Activity(description="hi"',
    ])
    with (
        self.assertRaises(mapping.MappingError),
        querying.track_queries() as queries
    ):
      querying.query('foo', Activity, lm=lm)
    self.assertEqual(len(queries), 1)
    self.assertTrue(pg.eq(queries[0].input, lf.Template('foo')))
    self.assertIn('foo', queries[0].lm_request.text)
    self.assertIn('Please respond to ', queries[0].lm_request.text)
    self.assertEqual(queries[0].schema.spec.cls, Activity)
    self.assertEqual(queries[0].default, lf.RAISE_IF_HAS_ERROR)
    self.assertTrue(queries[0].protocol.startswith('python:'))

    self.assertIsNone(queries[0].output)
    self.assertTrue(queries[0].has_error)
    self.assertTrue(queries[0].has_oop_error)
    self.assertIsNotNone(queries[0].error)
    self.assertIn('MappingError', queries[0].error.tag)
    self.assertIsNotNone(queries[0].end_time)
    self.assertIsNotNone(queries[0].usage_summary)

  def test_oop_error_for_query_with_schema_and_default(self):
    lm = fake.StaticSequence([
        'Activity(description="hi"',
    ])
    with querying.track_queries() as queries:
      querying.query('foo', Activity, default=1, lm=lm)

    self.assertEqual(len(queries), 1)
    self.assertTrue(pg.eq(queries[0].input, lf.Template('foo')))
    self.assertIn('foo', queries[0].lm_request.text)
    self.assertIn('Please respond to ', queries[0].lm_request.text)
    self.assertEqual(queries[0].schema.spec.cls, Activity)
    self.assertEqual(queries[0].default, 1)
    self.assertTrue(queries[0].protocol.startswith('python:'))

    self.assertEqual(queries[0].output, 1)
    self.assertTrue(queries[0].has_error)
    self.assertTrue(queries[0].has_oop_error)
    self.assertIsNotNone(queries[0].error)
    self.assertIn('MappingError', queries[0].error.tag)
    self.assertIsNotNone(queries[0].end_time)
    self.assertIsNotNone(queries[0].usage_summary)

  def test_lm_error_for_query_with_schema(self):
    class BadLLM(fake.StaticResponse):
      response = 'bad call'

      def _sample(self, *args, **kwargs):
        raise ValueError('bad LLM')

    with (
        self.assertRaises(ValueError),
        querying.track_queries() as queries
    ):
      querying.query('foo', Activity, default=1, lm=BadLLM())

    self.assertEqual(len(queries), 1)
    self.assertTrue(pg.eq(queries[0].input, lf.Template('foo')))
    self.assertIn('foo', queries[0].lm_request.text)
    self.assertIn('Please respond to ', queries[0].lm_request.text)
    self.assertEqual(queries[0].schema.spec.cls, Activity)
    self.assertEqual(queries[0].default, 1)
    self.assertTrue(queries[0].protocol.startswith('python:'))

    self.assertIsNone(queries[0].output)
    self.assertTrue(queries[0].has_error)
    self.assertIsNotNone(queries[0].error)
    self.assertIn('ValueError', queries[0].error.tag)
    self.assertIsNotNone(queries[0].end_time)
    self.assertIsNotNone(queries[0].usage_summary)

  def test_include_child_scopes(self):
    lm = fake.StaticSequence([
        'bar',
        'Activity(description="hi")',
    ])
    with querying.track_queries() as queries:
      querying.query('foo', lm=lm)
      with querying.track_queries() as child_queries:
        querying.query('give me an activity', Activity, lm=lm, examples=[
            mapping.MappingExample(input='2 + 2', schema=int, output=4)
        ])

    self.assertEqual(len(queries), 2)
    self.assertEqual(queries[0].lm_response, 'bar')
    self.assertTrue(pg.eq(queries[0].input, lf.Template('foo')))
    self.assertIsNone(queries[0].schema)
    self.assertEqual(queries[0].output, 'bar')
    self.assertIs(queries[0].lm, lm)

    self.assertTrue(pg.eq(queries[1].input, lf.Template('give me an activity')))
    self.assertEqual(queries[1].schema.spec.cls, Activity)
    self.assertIn('2 + 2', queries[1].lm_request.text)
    self.assertTrue(pg.eq(queries[1].output, Activity(description='hi')))
    self.assertTrue(pg.eq(queries[1].examples, [
        mapping.MappingExample(input='2 + 2', schema=int, output=4)
    ]))
    self.assertIs(queries[1].lm, lm)
    self.assertGreater(queries[0].elapse, 0)
    self.assertGreater(queries[0].usage_summary.total.total_tokens, 0)
    self.assertGreater(queries[1].usage_summary.total.total_tokens, 0)

    self.assertEqual(len(child_queries), 1)
    self.assertIs(child_queries[0], queries[1])

  def test_callbacks(self):
    lm = fake.StaticSequence([
        'bar',
    ])
    state = {}
    def start_callabck(query):
      self.assertFalse(query.is_completed)
      self.assertIsNone(query.end_time)
      elapse1 = query.elapse
      self.assertGreater(elapse1, 0)
      time.sleep(0.1)
      self.assertGreater(query.elapse, elapse1)
      self.assertIsNotNone(query.usage_summary)
      self.assertEqual(query.usage_summary.total.num_requests, 0)
      self.assertIsNone(query.lm_response)
      self.assertIsNone(query.output)
      self.assertIsNone(query.error)
      state['start'] = query

    def end_callback(query):
      self.assertTrue(query.is_completed)
      self.assertIsNotNone(query.end_time)
      self.assertIsNotNone(query.usage_summary)
      self.assertEqual(query.usage_summary.total.num_requests, 1)
      self.assertIsNotNone(query.lm_response)
      self.assertIsNotNone(query.output)
      self.assertIsNone(query.error)
      state['end'] = query

    with querying.track_queries(
        start_callabck=start_callabck, end_callabck=end_callback
    ) as queries:
      querying.query('foo', lm=lm)
    self.assertIs(state['start'], queries[0])
    self.assertIs(state['end'], queries[0])

  def test_callbacks_with_error(self):
    lm = fake.StaticSequence([
        'bar',
    ])
    state = {}
    def start_callabck(query):
      self.assertFalse(query.is_completed)
      self.assertIsNone(query.end_time)
      self.assertIsNotNone(query.usage_summary)
      self.assertEqual(query.usage_summary.total.num_requests, 0)
      self.assertIsNone(query.lm_response)
      self.assertIsNone(query.output)
      self.assertIsNone(query.error)
      state['start'] = query

    def end_callback(query):
      self.assertTrue(query.is_completed)
      self.assertIsNotNone(query.end_time)
      self.assertIsNotNone(query.usage_summary)
      self.assertEqual(query.usage_summary.total.num_requests, 1)
      self.assertIsNotNone(query.lm_response)
      self.assertIsNone(query.output)
      self.assertIsNotNone(query.error)
      state['end'] = query

    with self.assertRaises(mapping.MappingError):
      with querying.track_queries(
          start_callabck=start_callabck, end_callabck=end_callback
      ) as queries:
        querying.query('foo', int, lm=lm)
    self.assertIs(state['start'], queries[0])
    self.assertIs(state['end'], queries[0])

  def test_track_with_autofix(self):
    lm = fake.StaticSequence([
        '=1',
        inspect.cleandoc("""
            CorrectedCode(
                corrected_code='1',
            )
            """),
    ])
    with querying.track_queries() as queries:
      querying.query('foo', int, lm=lm, autofix=1)
    self.assertEqual(len(queries), 2)
    self.assertTrue(queries[0].has_error)
    self.assertTrue(queries[0].has_oop_error)
    self.assertIsNone(queries[0].output)
    self.assertFalse(queries[1].has_error)
    self.assertEqual(queries[1].output.__class__.__name__, 'CorrectedCode')
    self.assertEqual(queries[1].output.corrected_code, '1')

  def test_exclude_child_scopes(self):
    lm = fake.StaticSequence([
        'bar',
        'Activity(description="hi")',
    ])
    with querying.track_queries(include_child_scopes=False) as queries:
      querying.query('foo', lm=lm)
      with querying.track_queries(include_child_scopes=False) as child_queries:
        querying.query('give me an activity', Activity, lm=lm)

    self.assertEqual(len(queries), 1)
    self.assertTrue(pg.eq(queries[0].input, lf.Template('foo')))
    self.assertIsNone(queries[0].schema)
    self.assertEqual(queries[0].output, 'bar')
    self.assertIs(queries[0].lm, lm)

    self.assertEqual(len(child_queries), 1)
    self.assertTrue(
        pg.eq(child_queries[0].input, lf.Template('give me an activity'))
    )
    self.assertEqual(child_queries[0].schema.spec.cls, Activity)
    self.assertTrue(pg.eq(child_queries[0].output, Activity(description='hi')))
    self.assertIs(child_queries[0].lm, lm)

  def test_concurrent_map(self):

    def make_query(prompt):
      _ = querying.query(prompt, lm=lm)

    lm = fake.StaticSequence([
        'foo',
        'bar',
    ])
    with querying.track_queries() as queries:
      list(lf.concurrent_map(make_query, ['a', 'b']))
    self.assertEqual(len(queries), 2)


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""LangFunc test."""
import unittest

from langfun.core import component
from langfun.core import language_model
from langfun.core import message
from langfun.core import subscription
from langfun.core import template as template_lib
from langfun.core.langfunc import LangFunc
from langfun.core.langfunc import LangFuncCallEvent
from langfun.core.llms import fake
from langfun.core.llms.cache import in_memory
import pyglove as pg


class ExcitedEchoer(language_model.LanguageModel):
  """LM for testing."""

  def _sample(
      self,
      prompts: list[message.Message]) -> list[language_model.LMSamplingResult]:
    return [
        language_model.LMSamplingResult([
            language_model.LMSample(prompt.text + '!!!')
            ]) for prompt in prompts
    ]


class BasicTest(unittest.TestCase):

  def test_cached_lm_input_and_output(self):
    l = LangFunc('Hello')
    self.assertEqual(l.render(), 'Hello')
    self.assertIsNone(l.lm_input)
    with component.context(lm=ExcitedEchoer()):
      self.assertEqual(l(), 'Hello!!!')
      self.assertEqual(l.lm_input, 'Hello')
      self.assertEqual(l.lm_output, 'Hello!!!')

  def test_from_value(self):
    l1 = LangFunc.from_value('Hello')
    self.assertEqual(l1.template_str, 'Hello')

    l2 = LangFunc.from_value(l1)
    self.assertIs(l2, l1)

    l3 = LangFunc.from_value(l1, x=1)
    self.assertIsNot(l3, l1)
    self.assertTrue(pg.eq(l3, LangFunc('Hello', x=1)))

    c = template_lib.Template(
        '{{x}} + {{l}}',
        x=1,
        l=template_lib.Template('{{x}} + {{y}}', y=2))
    l3 = LangFunc.from_value(c.l)
    self.assertEqual(l3.render(), '1 + 2')

    l4 = LangFunc.from_value(1)
    self.assertEqual(l4.render(), '1')


class LangFuncCallTest(unittest.TestCase):

  def test_call(self):
    l = LangFunc('Hello', lm=ExcitedEchoer())
    self.assertEqual(l, 'Hello')
    self.assertNotEqual(l, 'Hello!!!')
    self.assertEqual(l.natural_language_format(), 'Hello')

    i = l.render()
    self.assertEqual(i, 'Hello')
    self.assertEqual(i, message.UserMessage('Hello'))
    self.assertEqual(i.tags, ['rendered'])

    r = l()
    self.assertEqual(
        r,
        message.AIMessage(
            'Hello!!!', score=0.0, logprobs=None, is_cached=False,
            usage=language_model.UsageNotAvailable()
        )
    )
    self.assertEqual(r.tags, ['lm-response', 'lm-output'])
    self.assertEqual(
        r.source,
        message.UserMessage('Hello', metadata=dict(cache_seed=0))
    )
    self.assertEqual(r.source.tags, ['rendered', 'lm-input'])

    self.assertEqual(str(l), 'Hello')
    self.assertEqual(
        repr(l),
        "LangFunc(template_str='Hello', clean=True,"
        ' lm=ExcitedEchoer(sampling_options=LMSamplingOptions(temperature=None,'
        ' max_tokens=None, n=1, top_k=40, top_p=None, stop=None,'
        ' random_seed=None, logprobs=False, top_logprobs=None,'
        ' max_thinking_tokens=None, reasoning_effort=None), cache=None,'
        ' max_concurrency=None, timeout=120.0, max_attempts=5,'
        ' retry_interval=(5, 60), exponential_backoff=True,'
        ' max_retry_interval=300, debug=False))',
    )

    l = LangFunc('Hello')
    with component.context(lm=ExcitedEchoer()):
      self.assertEqual(l, 'Hello')
      self.assertEqual(l.natural_language_format(), 'Hello')
      self.assertEqual(l.render(), 'Hello')
      r = l(cache_seed=1)
      self.assertEqual(
          r,
          message.AIMessage(
              'Hello!!!', score=0.0, logprobs=None, is_cached=False,
              usage=language_model.UsageNotAvailable()
          )
      )
      self.assertEqual(r.tags, ['lm-response', 'lm-output'])
      self.assertEqual(r.source.metadata.cache_seed, 1)

    self.assertEqual(str(l), 'Hello')

    with self.assertRaisesRegex(
        AttributeError, '`lm` is not found under its context'
    ):
      l()

  def test_call_with_init_time_vars(self):
    l = LangFunc('Hello, {{foo}}', foo='a')
    self.assertEqual(l.foo, 'a')
    self.assertEqual(l.natural_language_format(), 'Hello, a')

    with self.assertRaises(AttributeError):
      _ = l.bar
    self.assertEqual(l.render(), 'Hello, a')

    # Use instance variable even the parent has the same variable.
    r = pg.Dict(s=LangFunc('Hello, {{x}}', x='foo'), x='world')
    self.assertEqual(r.s.x, 'foo')
    self.assertEqual(r.x, 'world')
    self.assertEqual(r.s.render(), 'Hello, foo')

  def test_call_with_call_time_vars(self):
    l = LangFunc('Hello, {{foo}} {{bar}}', bar=LangFunc('and {{foo}}'))
    with self.assertRaises(AttributeError):
      _ = l.foo
    self.assertEqual(l.render(foo='a'), 'Hello, a and a')

    l = LangFunc('Hello, {{foo}} {{bar}}',
                 foo='a',
                 bar=LangFunc('and {{foo}}', foo='b'))
    self.assertEqual(l.render(), 'Hello, a and b')
    self.assertEqual(l.render(foo='c'), 'Hello, c and b')

  def test_call_with_contextual_vars(self):
    class Chat(LangFunc):
      """Chat.

      {{conversation}}
      """

      # Generated texts.
      prompt: str = LangFunc('compute {{x}} + {{y}}')
      response: str = LangFunc('{{x + y}}')
      conversation: str = LangFunc('Q: {{prompt}}\n\nA: {{response}}')

      # Instance variables.
      x: int = 1

      # Class variables.
      y = 2

    c = Chat()
    self.assertEqual(c.prompt.x, 1)
    self.assertEqual(c.prompt.y, 2)
    self.assertEqual(c.response.x, 1)
    self.assertEqual(c.response.y, 2)
    self.assertIs(c.conversation.prompt, c.prompt)
    self.assertIs(c.conversation.response, c.response)

    self.assertEqual(c.render(), 'Q: compute 1 + 2\n\nA: 3')
    self.assertEqual(c.render(x=2), 'Q: compute 2 + 2\n\nA: 4')

    c = Chat(x=2)
    self.assertEqual(c.render(), 'Q: compute 2 + 2\n\nA: 4')
    self.assertEqual(c.render(y=3), 'Q: compute 2 + 3\n\nA: 5')

  def test_call_with_overriden_lm_input(self):
    t = LangFunc('Hello')
    with component.context(lm=ExcitedEchoer()):
      self.assertEqual(t(lm_input=message.UserMessage('Hi')), 'Hi!!!')

  def test_call_with_cache(self):
    l = LangFunc('{{x}}')
    with component.context(
        lm=fake.StaticSequence(['a', 'b', 'c', 'd'], cache=in_memory.InMemory())
    ):
      self.assertEqual(l(x=1), 'a')
      self.assertEqual(l(x=2), 'b')
      self.assertEqual(l(x=1), 'a')
      self.assertEqual(l(x=1, cache_seed=None), 'c')
      self.assertEqual(l(x=1, cache_seed=None), 'd')
      self.assertEqual(l(x=2), 'b')

  def test_call_with_skip_lm(self):
    l = LangFunc('hi')
    with component.context(lm=ExcitedEchoer()):
      self.assertEqual(l(skip_lm=True), 'hi')


class CallEventTest(unittest.TestCase):

  def test_call_event(self):
    l = LangFunc(
        'Subject: {{subject}}',
        subject=LangFunc(
            'The science of {{ask_me}}',
            ask_me=LangFunc(
                '{{inner_voice()}}', inner_voice=LangFunc('living {{name}}')
            ),
        ),
    )

    class CallEventHandler(subscription.EventHandler[LangFuncCallEvent]):

      def __init__(self):
        super().__init__()
        self.call_events = []
        self.call_stacks = []

      def on_event(self, event: LangFuncCallEvent):
        self.call_events.append((event.lm_input, event.lm_output.text))
        self.call_stacks.append(event.lm_callstack)

    x = CallEventHandler()
    subscription.subscribe(x)
    with component.context(lm=ExcitedEchoer()):
      l(name='spirit')

    self.assertEqual(
        x.call_events,
        [
            ('living spirit', 'living spirit!!!'),
            (
                'Subject: The science of living spirit!!!',
                'Subject: The science of living spirit!!!!!!',
            ),
        ],
    )
    self.assertEqual(x.call_stacks, [[l], []])
    subscription.unsubscribe(x)

    y = CallEventHandler()
    z = CallEventHandler()
    subscription.subscribe(y, l.subject)
    subscription.subscribe(z, l.subject.ask_me.inner_voice)

    with component.context(lm=ExcitedEchoer()):
      l(name='dead')

    # l.subject is not called explicitly.
    self.assertEqual(len(y.call_events), 0)

    # l.subject.ask_me.inner_voice is called explicitly.
    self.assertEqual(
        z.call_events,
        [
            ('living dead', 'living dead!!!'),
        ],
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Demonstrations."""

from typing import Annotated

import langfun.core as lf
from langfun.core.templates.completion import Completion


@lf.use_init_args(['examples', 'description'])
class Demonstration(lf.LangFunc):
  """Base class for demonstration.

  {{ description }}
  {% for example in examples %}
    Example {{loop.index}}:
  {{ example.render().text | indent(4, True) }}
  {% endfor %}
  """

  examples: Annotated[
      list[Completion],
      (
          'A list of completions as few-shot examples.'
      )
  ]

  description: Annotated[
      lf.LangFunc,
      (
          'A leading description of the demonstrations.'
      )
  ] = lf.LangFunc('Here are the examples for demonstrating this:')

  def __bool__(self):
    return bool(self.examples)

  def __call__(self, *args, **kwargs):
    raise ValueError(f'{self.__class__.__name__} is not callable.')



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for natural language utilities."""
import unittest

from langfun.core import component
from langfun.core import natural_language as nl
import pyglove as pg


class NaturalLanguageFormattableTest(unittest.TestCase):
  """Tests for lf.NaturalLanguageFormattable."""

  def test_natural_language_formattable(self):
    class A(nl.NaturalLanguageFormattable, component.Component):
      x: int
      y: str

      def natural_language_format(self) -> str:
        return f'A simple object with {self.x} and {self.y!r}.'

    a = A(1, 'abc')
    self.assertEqual(repr(a), 'A(x=1, y=\'abc\')')
    with pg.object_utils.repr_format(natural_language=True):
      self.assertEqual(repr(a), "A simple object with 1 and 'abc'.")
    self.assertEqual(str(a), "A simple object with 1 and 'abc'.")


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Langfun agentic framework.."""

# pylint: disable=g-bad-import-order
# pylint: disable=g-importing-member
# pylint: disable=g-import-not-at-top

from langfun.core.agentic.action import ActionError
from langfun.core.agentic.action import ActionTimeoutError

from langfun.core.agentic.action import Action
from langfun.core.agentic.action import ActionInvocation
from langfun.core.agentic.action import ExecutionTrace
from langfun.core.agentic.action import ParallelExecutions
from langfun.core.agentic.action import Session

from langfun.core.agentic.action_eval import ActionEval
from langfun.core.agentic.action_eval import ActionEvalV1


# pylint: enable=g-bad-import-order
# pylint: enable=g-importing-member
# pylint: enable=g-import-not-at-top



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for OpenAI models."""

import unittest
import langfun.core as lf
from langfun.core.llms import openai


class OpenAITest(unittest.TestCase):
  """Tests for OpenAI language model."""

  def test_dir(self):
    self.assertIn('gpt-4-turbo', openai.OpenAI.dir())

  def test_key(self):
    with self.assertRaisesRegex(ValueError, 'Please specify `api_key`'):
      openai.Gpt4()('hi')

  def test_model_id(self):
    self.assertEqual(
        openai.Gpt35(api_key='test_key').model_id, 'text-davinci-003')

  def test_resource_id(self):
    self.assertEqual(
        openai.Gpt35(api_key='test_key').resource_id,
        'openai://text-davinci-003'
    )

  def test_headers(self):
    self.assertEqual(
        openai.Gpt35(api_key='test_key').headers,
        {
            'Content-Type': 'application/json',
            'Authorization': 'Bearer test_key',
        },
    )

  def test_max_concurrency(self):
    self.assertGreater(
        openai.Gpt4o(api_key='test_key').max_concurrency, 0
    )

  def test_request_args(self):
    self.assertEqual(
        openai.Gpt4(api_key='test_key')._request_args(
            lf.LMSamplingOptions(
                temperature=1.0, stop=['\n'], n=1, random_seed=123
            )
        ),
        dict(
            model='gpt-4',
            top_logprobs=None,
            n=1,
            temperature=1.0,
            stop=['\n'],
            seed=123,
        ),
    )
    with self.assertRaisesRegex(RuntimeError, '`logprobs` is not supported.*'):
      openai.GptO1Preview(api_key='test_key')._request_args(
          lf.LMSamplingOptions(
              temperature=1.0, logprobs=True
          )
      )

  def test_estimate_cost(self):
    self.assertEqual(
        openai.Gpt4(api_key='test_key').estimate_cost(
            lf.LMSamplingUsage(
                total_tokens=200,
                prompt_tokens=100,
                completion_tokens=100,
            )
        ),
        0.009
    )

  def test_lm_get(self):
    self.assertIsInstance(lf.LanguageModel.get('gpt-4o'), openai.OpenAI)


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Image tests."""
import io
import unittest
from unittest import mock

import langfun.core as lf
from langfun.core.modalities import image as image_lib
from langfun.core.modalities import mime as mime_lib
import PIL.Image as pil_image
import pyglove as pg


image_content = (
    b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x18\x00\x00\x00\x18\x04'
    b'\x03\x00\x00\x00\x12Y \xcb\x00\x00\x00\x18PLTE\x00\x00'
    b'\x00fff_chaag_cg_ch^ci_ciC\xedb\x94\x00\x00\x00\x08tRNS'
    b'\x00\n\x9f*\xd4\xff_\xf4\xe4\x8b\xf3a\x00\x00\x00>IDATx'
    b'\x01c \x05\x08)"\xd8\xcc\xae!\x06pNz\x88k\x19\\Q\xa8"\x10'
    b'\xc1\x14\x95\x01%\xc1\n\xa143Ta\xa8"D-\x84\x03QM\x98\xc3'
    b'\x1a\x1a\x1a@5\x0e\x04\xa0q\x88\x05\x00\x07\xf8\x18\xf9'
    b'\xdao\xd0|\x00\x00\x00\x00IEND\xaeB`\x82'
)


def mock_request(*args, **kwargs):
  del args, kwargs
  return pg.Dict(content=image_content)


class ImageTest(unittest.TestCase):

  def test_from_bytes(self):
    image = image_lib.Image.from_bytes(image_content)
    self.assertEqual(image.image_format, 'png')
    self.assertIn('data:image/png;base64,', image._raw_html())
    self.assertEqual(image.to_bytes(), image_content)
    with self.assertRaisesRegex(
        lf.ModalityError, '.* cannot be converted to text'
    ):
      image.to_text()

  def test_from_bytes_invalid(self):
    image = image_lib.Image.from_bytes(b'bad')
    with self.assertRaisesRegex(ValueError, 'Expected MIME type'):
      _ = image.image_format

  def test_from_bytes_base_cls(self):
    self.assertIsInstance(
        mime_lib.Mime.from_bytes(image_content), image_lib.Image
    )

  def test_from_uri(self):
    image = image_lib.Image.from_uri('http://mock/web/a.png')
    with mock.patch('requests.get') as mock_requests_get:
      mock_requests_get.side_effect = mock_request
      self.assertEqual(image.image_format, 'png')
      self.assertEqual(
          image._raw_html(),
          '<img src="http://mock/web/a.png">'
      )
      self.assertEqual(image.to_bytes(), image_content)

  def test_from_uri_base_cls(self):
    with mock.patch('requests.get') as mock_requests_get:
      mock_requests_get.side_effect = mock_request
      image = mime_lib.Mime.from_uri('http://mock/web/a.png')
      self.assertIsInstance(image, image_lib.Image)
      self.assertEqual(image.image_format, 'png')
      self.assertEqual(
          image._raw_html(),
          '<img src="http://mock/web/a.png">'
      )
      self.assertEqual(image.to_bytes(), image_content)

  def test_image_size(self):
    image = image_lib.Image.from_uri('http://mock/web/a.png')
    with mock.patch('requests.get') as mock_requests_get:
      mock_requests_get.side_effect = mock_request
      self.assertEqual(image.size, (24, 24))

  def test_to_pil_image(self):
    image = image_lib.Image.from_uri('http://mock/web/a.png')
    with mock.patch('requests.get') as mock_requests_get:
      mock_requests_get.side_effect = mock_request
      self.assertIsInstance(image.to_pil_image(), pil_image.Image)

  def test_from_pil_image(self):
    image = pil_image.open(io.BytesIO(image_content))
    self.assertIsInstance(
        image_lib.Image.from_pil_image(image), image_lib.Image
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""MIME type data."""

import base64
import functools
from typing import Annotated, Any, Iterable, Type, Union
import langfun.core as lf
# Placeholder for Google-internal internet access import.
import puremagic as pm
import pyglove as pg
import requests   # pylint: disable=unused-import


def _detect_mime_type(content: bytes) -> str:
  """Returns the MIME type from the given content."""
  try:
    return pm.from_string(content, mime=True).lower()
  except pm.PureError:
    try:
      content.decode('utf-8')
      return 'text/plain'
    except UnicodeDecodeError:
      return 'application/octet-stream'


class Mime(lf.Modality):
  """Base for MIME data."""

  # The regular expression that describes the MIME type str.
  # If None, the MIME type is dynamic. Subclass could override.
  MIME_PREFIX = None

  uri: Annotated[str | None, 'The URI for locating the MIME data. '] = None

  content: Annotated[
      Union[str, bytes, None], 'The raw content of the MIME type.'
  ] = None

  @functools.cached_property
  def mime_type(self) -> str:
    """Returns the MIME type."""
    mime = _detect_mime_type(self.to_bytes())
    if (
        self.MIME_PREFIX
        and not mime.lower().startswith(self.MIME_PREFIX)
        # NOTE(daiyip): libmagic fails to detect the MIME type of some binary
        # files.
        and mime != 'application/octet-stream'
    ):
      raise ValueError(
          f'Expected MIME type: {self.MIME_PREFIX}, Encountered: {mime}'
      )
    return mime

  @functools.cached_property
  def is_text(self) -> bool:
    return self.mime_type.startswith((
        'text/',
        'application/javascript',
        'application/json',
        'application/ld+json',
        'application/plain',
        'application/rtf',
        'application/xhtml+xml',
        'application/xml',
        'application/x-javascript',
        'application/x-python-code',
        'application/x-tex',
        'application/x-typescript',
        'application/x-yaml',
    ))

  @property
  def is_binary(self) -> bool:
    """Returns True if the MIME type is a binary type."""
    return not self.is_text

  def to_text(self) -> str:
    """Returns the text content of the MIME type."""
    if not self.is_text:
      raise lf.ModalityError(
          f'MIME type {self.mime_type!r} cannot be converted to text.'
      )
    return self.to_bytes().decode()

  def is_compatible(
      self, mime_types: str | Iterable[str]
  ) -> bool:
    """Returns True if this object is compatible to any of the MIME types."""
    if isinstance(mime_types, str):
      mime_types = {mime_types}
    return self._is_compatible(mime_types)

  def _is_compatible(self, mime_types: Iterable[str]):
    if not mime_types:
      return False
    return self.mime_type in mime_types

  def make_compatible(
      self,
      mime_types: str | Iterable[str]
      ) -> Union['Mime', list['Mime']]:
    """Makes compatible MIME objects from this object."""
    if isinstance(mime_types, str):
      mime_types = {mime_types}
    if not self._is_compatible(mime_types):
      raise lf.ModalityError(
          f'MIME type {self.mime_type!r} cannot be converted to supported '
          f'types: {mime_types!r}.'
      )
    return self._make_compatible(mime_types)

  def _make_compatible(
      self,
      mime_types: Iterable[str]
  ) -> Union['Mime', list['Mime']]:
    """Makes compatbile MIME objects from this object."""
    del mime_types
    return self

  def _on_bound(self):
    super()._on_bound()
    if self.uri is None and self.content is None:
      raise ValueError('Either uri or content must be provided.')

  def to_bytes(self) -> bytes:
    if self.content is not None:
      return self.content

    self.rebind(content=self.download(self.uri), skip_notification=True)
    return self.content

  @property
  def content_uri(self) -> str:
    """Returns the URI with encoded content."""
    base64_content = base64.b64encode(self.to_bytes()).decode()
    return f'data:{self.mime_type};base64,{base64_content}'

  @property
  def embeddable_uri(self) -> str:
    """Returns the URI that can be embedded in HTML."""
    if self.uri and self.uri.lower().startswith(('http:', 'https:', 'ftp:')):
      return self.uri
    return self.content_uri

  @classmethod
  def from_uri(cls, uri: str, **kwargs) -> 'Mime':
    if uri.startswith('data:'):
      mime_type, content = cls._parse_data_uri(uri)
      return cls.class_from_mime_type(mime_type).from_bytes(content, **kwargs)

    if cls is Mime:
      content = cls.download(uri)
      mime = _detect_mime_type(content)
      return cls.class_from_mime_type(mime)(uri=uri, content=content, **kwargs)
    return cls(uri=uri, content=None, **kwargs)

  @classmethod
  def _parse_data_uri(cls, uri: str) -> tuple[str, bytes]:
    """Returns the MIME type and content from the given data URI."""
    assert uri.startswith('data:'), uri
    mime_end_pos = uri.find(';', 0)
    if mime_end_pos == -1:
      raise ValueError(f'Invalid data URI: {uri!r}.')
    mime_type = uri[5: mime_end_pos].strip().lower()
    encoding_end_pos = uri.find(',', mime_end_pos + 1)
    if encoding_end_pos == -1:
      raise ValueError(f'Invalid data URI: {uri!r}.')
    encoding = uri[mime_end_pos + 1: encoding_end_pos].strip().lower()
    if encoding != 'base64':
      raise ValueError(f'Unsupported encoding: {encoding!r}.')
    base64_content = uri[encoding_end_pos + 1:].strip().encode()
    return mime_type, base64.b64decode(base64_content)

  @classmethod
  def from_bytes(cls, content: bytes | str, **kwargs) -> 'Mime':
    if cls is Mime:
      return cls.class_from_mime_type(
          _detect_mime_type(content)
      )(content=content, **kwargs)
    return cls(content=content, **kwargs)

  @classmethod
  def class_from_mime_type(cls, mime_type: str) -> Type['Mime']:
    """Subclass from the given MIME type."""
    for subcls in cls.__subclasses__():
      if subcls.MIME_PREFIX is not None and mime_type.startswith(
          subcls.MIME_PREFIX):
        return subcls
    return cls

  @classmethod
  def download(cls, uri: str) -> bytes | str:
    """Downloads the content of the given URI."""
    if uri.lower().startswith(('http:', 'https:', 'ftp:')):
      return requests.get(uri, headers={'User-Agent': 'Mozilla/5.0'}).content
    else:
      content = pg.io.readfile(uri, mode='rb')
      assert content is not None
      return content

  def _html_tree_view_content(
      self,
      **kwargs) -> str:
    return self._raw_html()

  def _html_tree_view(
      self,
      view: pg.views.HtmlTreeView,
      extra_flags: dict[str, Any] | None = None,
      **kwargs
  ):
    extra_flags = extra_flags if extra_flags is not None else {}
    raw_mime_content = extra_flags.get('raw_mime_content', False)
    display_modality_when_hover = extra_flags.get(
        'display_modality_when_hover', False
    )
    if raw_mime_content:
      kwargs['enable_summary'] = False
    elif display_modality_when_hover:
      kwargs.update(
          enable_summary=True,
          enable_summary_tooltip=True,
      )
    return super()._html_tree_view(
        view=view, extra_flags=extra_flags, **kwargs
    )

  def _html_tree_view_summary(
      self,
      *,
      view: pg.views.HtmlTreeView,
      extra_flags: dict[str, Any] | None = None,
      **kwargs
  ):
    extra_flags = extra_flags or {}
    if extra_flags.get('display_modality_when_hover', False):
      def summary_tooltip(*args, content: str | None = None, **kwargs):
        del content
        return view.tooltip(*args, content=self._raw_html(), **kwargs)
    else:
      summary_tooltip = None
    return super()._html_tree_view_summary(
        view=view,
        summary_tooltip_fn=summary_tooltip,
        extra_flags=extra_flags,
        **kwargs
    )

  def _raw_html(self) -> str:
    if self.uri and self.uri.lower().startswith(('http:', 'https:', 'ftp:')):
      uri = self.uri
    else:
      uri = self.content_uri
    return self._mime_control_for(uri)

  def _mime_control_for(self, uri) -> str:
    return f'<embed type="{self.mime_type}" src="{uri}"/>'


@pg.use_init_args(['mime', 'content', 'uri'])
class Custom(Mime):
  """Custom MIME data."""

  mime: Annotated[
      str, 'The MIME type of the data. E.g. text/plain, or image/png. '
  ]

  @property
  def mime_type(self) -> str:
    return self.mime



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any
import unittest

from langfun.core import modality
import pyglove as pg


class CustomModality(modality.Modality):
  content: str

  def to_bytes(self):
    return self.content.encode()


class ModalityTest(unittest.TestCase):

  def test_basic(self):
    v = CustomModality('a')
    self.assertIsNone(v.referred_name)
    self.assertEqual(str(v), "CustomModality(\n  content = 'a'\n)")
    self.assertEqual(v.hash, '0cc175b9')

    _ = pg.Dict(metadata=pg.Dict(x=pg.Dict(metadata=pg.Dict(y=v))))
    self.assertEqual(v.referred_name, 'x.metadata.y')
    self.assertEqual(str(v), "CustomModality(\n  content = 'a'\n)")
    with modality.format_modality_as_ref():
      self.assertEqual(str(v), '<<[[x.metadata.y]]>>')


class ModalityRefTest(unittest.TestCase):

  def test_placehold(self):
    class A(pg.Object):
      x: Any
      y: Any

    a = A(x=dict(z=CustomModality('a')), y=CustomModality('b'))
    self.assertEqual(
        modality.ModalityRef.placehold(a),
        A(x=dict(z=modality.ModalityRef('x.z')), y=modality.ModalityRef('y')),
    )
    self.assertEqual(
        modality.ModalityRef.placehold(a.x),
        # The prefix 'x' of referred name is preserved.
        dict(z=modality.ModalityRef('x.z')),
    )

  def test_from_value(self):
    class A(pg.Object):
      x: Any
      y: Any

    a = A(x=dict(z=CustomModality('a')), y=CustomModality('b'))
    self.assertTrue(
        pg.eq(
            modality.Modality.from_value(a),
            {
                'x.z': CustomModality('a'),
                'y': CustomModality('b'),
            },
        )
    )
    self.assertTrue(
        pg.eq(
            modality.Modality.from_value(a.x.z),
            {
                'x.z': CustomModality('a'),
            },
        )
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Console utilities."""

import sys
from typing import Any
import pyglove as pg


def write(
    value: Any,
    *,
    title: str | None = None,
    color: str | None = None,
    background: str | None = None,
    styles: list[str] | None = None
) -> None:
  """Writes text to console.

  Args:
    value: A Python value to write to console. If not a string. `str(value)`
      will be called as the text to write.
    title: An optional title printed in bold.
    color: A string for text colors. Applicable values are: 'red', 'green',
      'yellow', 'blue', 'magenta', 'cyan', 'white'.
    background: A string for background colors. Applicable values are: 'red',
      'green', 'yellow', 'blue', 'magenta', 'cyan', 'white'.
    styles: A list of strings for applying styles on the text.
      Applicable values are: 'bold', 'dark', 'underline', 'blink', 'reverse',
      'concealed'.
  """
  # Print title if present.
  if title is not None:
    print(pg.colored(title, styles=['bold']))

  # Print body.
  print(
      pg.colored(
          str(value), color=color, background=background, styles=styles
      )
  )


_notebook = None
try:
  ipython_module = sys.modules['IPython']
  if 'IPKernelApp' in ipython_module.get_ipython().config:
    _notebook = ipython_module.display
except (KeyError, AttributeError):  # pylint: disable=broad-except
  pass


def under_notebook() -> bool:
  """Returns True if current process runs under notebook."""
  return bool(_notebook)


def display(value: Any, clear: bool = False) -> Any:  # pylint: disable=redefined-outer-name
  """Displays object in current notebook cell."""
  if _notebook is not None:
    if clear:
      _notebook.clear_output()
    return _notebook.display(value)
  return None


def run_script(javascript: str) -> Any:
  """Runs JavaScript in current notebook cell."""
  if _notebook is not None:
    return _notebook.display(_notebook.Javascript(javascript))
  return


def clear() -> None:
  """Clears output from current notebook cell."""
  if _notebook is not None:
    _notebook.clear_output()



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Scoring evaluation."""

import abc
import io
import os
from typing import Any
import langfun.core as lf
from langfun.core.eval import base
import pyglove as pg


class Scoring(base.Evaluation):
  """Base class of evaluations by scoring."""

  # CONSTANTS.
  SCORED_JSON = 'scored.json'
  SCORED_HTML = 'scored.html'

  @property
  def scored(self) -> list[tuple[Any, Any, float, lf.Message]]:
    """Returns a list of (example, structured output, score, output message)."""
    return self._scored

  @property
  def num_scored(self) -> int:
    """Returns the number of scored examples."""
    return len(self._scored)

  @property
  def score_rate(self) -> float:
    """Returns the score rate."""
    if self.num_completed == 0:
      return 0.0
    return self.num_scored / self.num_completed

  @property
  def scored_link(self) -> str:
    """Returns the matches page."""
    return self.link(os.path.join(self.dir, Scoring.SCORED_HTML))

  @property
  def avg_score(self) -> float:
    if self.num_scored == 0:
      return 0
    return sum([i[2] for i in self._scored]) / self.num_scored

  def _reset(self) -> None:
    super()._reset()
    self._scored = []

  def audit_processed(
      self, example_idx: int, example: Any, output: Any, message: lf.Message,
      dryrun: bool = False
  ) -> None:
    del example_idx
    score = self.score(example, output)

    if dryrun:
      lf.console.write('')
      lf.console.write(
          str(score),
          title='SCORE',
          color='blue',
      )
    self._scored.append((example, output, score, message))

  @abc.abstractmethod
  def score(self, example: Any, output: Any) -> float:
    """Scores the output against its input example."""

  def _eval_status(self, progress: lf.concurrent.Progress) -> dict[str, Any]:
    del progress
    return {
        'Average Score': {self.avg_score},
        'Scored': '%.3f%% (%d/%d)' % (
            self.score_rate * 100,
            self.num_scored,
            self.num_completed,
        ),
        'Failed': '%.3f%% (%d/%d)' % (
            self.failure_rate * 100,
            self.num_failures,
            self.num_completed,
        ),
    }

  def _completion_status(self, run_status: str) -> str:
    assert self.result is not None
    m = self.result.metrics
    return (
        'COMPLETED(%s): AvgScore=%f Scored=%.3f%% (%d/%d) '
        'Failures=%.3f%% (%d/%d)'
    ) % (
        run_status,
        m.avg_score,
        m.score_rate * 100,
        m.num_scored,
        m.total,
        m.failure_rate * 100,
        m.failures,
        m.total,
    )

  def finalize(self) -> pg.Dict:
    result = super().finalize()
    result.metrics.update(
        num_scored=self.num_scored,
        score_rate=self.score_rate,
        avg_score=self.avg_score,
    )
    return result

  def save(
      self, definition: bool = True, result: bool = True, report: bool = True
  ) -> None:
    super().save(definition, result, report)

    if result:
      # Save scored.
      pg.save(
          [
              # We force the output to be dict as its type may be defined
              # within functors which could be deserialized.
              pg.Dict(input=input, output=output, score=score)
              for input, output, score, _ in self.scored
          ],
          os.path.join(self.dir, Scoring.SCORED_JSON),
      )

    if report:
      pg.save(
          self._html([self._render_result, self._render_scored]),
          os.path.join(self.dir, Scoring.SCORED_HTML),
          file_format='txt',
      )

  def _render_result_header(self, s: io.StringIO):
    super()._render_result_header(s)
    s.write('<td>Avg Score</td>')
    s.write('<td>Scored</td>')

  def _render_result_row(self, s: io.StringIO):
    super()._render_result_row(s)
    s.write(
        '<td><span style="color:blue">%.3f</span></td>' % self.avg_score
    )
    s.write(
        '<td><span style="color:red">%s</span>%s</td>'
        % (
            '%.3f%% ' % (self.score_rate * 100),
            '<a href="%s">(%d/%d)</a>'
            % (self.scored_link, self.num_scored, self.num_completed),
        )
    )

  def _render_summary_metrics(self, s: io.StringIO) -> None:
    """Renders metrics in HTML."""
    assert self.result is not None
    m = self.result.metrics
    self._render_link(
        s,
        'Average score (%d/%d)' % (m.num_scored, m.total),
        '%.3f (%.3f%%)' % (m.avg_score, m.score_rate * 100),
        'color:green',
        lambda: self.scored_link,
    )
    s.write(' | ')
    super()._render_summary_metrics(s)

  def _render_scored(self, s: io.StringIO) -> None:
    """Formats the matched cases into html."""
    s.write('<h2> Scored </h2>')
    s.write('<div style="white-space:pre">\n')
    s.write(
        '<table style="border: 1px solid;">'
        '<tr class="header">'
        '<td>No.</td><td>Input</td><td>Output</td><td>Score</td>'
        '<td>Prompt/Response Chain</td>'
        '</tr>'
    )
    for i, (example, output, score, message) in enumerate(self.scored):
      bgcolor = 'white' if i % 2 == 0 else '#DDDDDD'
      s.write(f'<tr style="background-color: {bgcolor}"><td>{i + 1}</td>')
      input_str = pg.Html.escape(
          pg.format(example, verbose=False, max_bytes_len=32)
      )
      s.write(f'<td style="color:green;white-space:pre-wrap">{input_str}</td>')
      output_str = pg.Html.escape(
          pg.format(output, verbose=False, max_bytes_len=32)
      )
      s.write(f'<td style="color:blue;white-space:pre-wrap">{output_str}</td>')
      s.write(f'<td style="color:magenta;white-space:pre-wrap">{score}</td>')
      s.write('<td>')
      self._render_message(message, s)
      s.write('</td>')
      s.write('</tr>')
    s.write('</table></div>')



# Copyright 2023 The Langfun Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""LM cache base."""

import abc
import dataclasses
import datetime
from typing import Annotated, Any, Callable
import langfun.core as lf


@dataclasses.dataclass(frozen=True)
class LMCacheEntry:
  """LM cache entry."""
  result: lf.LMSamplingResult
  expire: datetime.datetime | None = None


class LMCacheBase(lf.LMCache):
  """The common LMCache base."""

  key: Annotated[
      Callable[[lf.LanguageModel, lf.Message, int], Any] | None,
      (
          'A callable ojbect used for computing the key (hashable structure) '
          'from the language model used, input prompt and a cache seed. '
          'If None, a default key will be used, which will be determiend by '
          'model id, sampling options, input prompt and the cache seed.'
      ),
  ] = None

  ttl: Annotated[
      int | None,
      (
          'Time-to-live in seconds.'
      )
  ] = None

  def _on_bound(self):
    super()._on_bound()
    self._key = self.key or default_key
    self._stats = lf.LMCache.Stats()

  @property
  def stats(self) -> lf.LMCache.Stats:
    return self._stats

  def get(
      self, lm: lf.LanguageModel, prompt: lf.Message, seed: int
  ) -> lf.LMSamplingResult | None:
    """Gets the cached result of a prompt generated by a language model."""
    key = self._key(lm, prompt, seed)
    entry = self._get(lm.model_id, key)
    self._stats.num_queries += 1
    if entry is None:
      self._stats.num_misses += 1
      return None
    if entry.expire is not None and entry.expire < datetime.datetime.now():
      self._stats.num_hit_expires += 1
      self._stats.num_deletes += 1
      assert self._delete(lm.model_id, key)
      return None
    self._stats.num_hits += 1
    return entry.result

  def put(
      self,
      lm: lf.LanguageModel,
      prompt: lf.Message,
      result: lf.LMSamplingResult,
      seed: int,
  ) -> None:
    """Puts the result of a prompt generated by a language model in cache."""
    expire = None
    if self.ttl:
      expire = datetime.datetime.now() + datetime.timedelta(seconds=self.ttl)
    entry = LMCacheEntry(result, expire)
    self._put(lm.model_id, self._key(lm, prompt, seed), entry)
    self._stats.num_updates += 1

  def delete(
      self,
      lm: lf.LanguageModel,
      prompt: lf.Message,
      seed: int,
  ) -> bool:
    """Deletes the result of a prompt generated by a language model in cache."""
    deleted = self._delete(lm.model_id, self._key(lm, prompt, seed))
    if deleted:
      self._stats.num_deletes += 1
    return deleted

  @abc.abstractmethod
  def _get(self, model_id: str, key: str) -> LMCacheEntry | None:
    """Returns a LM cache entry associated with the key."""

  @abc.abstractmethod
  def _put(self, model_id: str, key: str, entry: LMCacheEntry) -> None:
    """Puts a LM cache entry associated with the key."""

  @abc.abstractmethod
  def _delete(self, model_id: str, key: str) -> bool:
    """Deletes a LM cache entry associated with the key."""

  def _sym_clone(self, deep: bool, memo: Any = None) -> 'LMCacheBase':
    v = super()._sym_clone(deep, memo)
    v._stats = self._stats  # pylint: disable=protected-access
    return v


def default_key(lm: lf.LanguageModel, prompt: lf.Message, seed: int) -> Any:
  """Default key for LM cache."""
  return (prompt.text_with_modality_hash, lm.sampling_options.cache_key(), seed)
